# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. |JtagContext| replace::   :class:`JtagContext`

Jtag service provides basic operations to run Jtag-based transactions on a target.


.. _properties:

Properties
^^^^^^^^^^
+--------------------------+-------+------------------------------------------+
| Name                     | Type  | Description                              |
+==========================+=======+==========================================+
| ID                       | |str| | Context ID of the node in the hw_server  |
+--------------------------+-------+------------------------------------------+
| ParentID                 | |str| | Context ID of the parent node in the     |
|                          |       | hw_server.                               |
+--------------------------+-------+------------------------------------------+
| Name                     | |str| | Name of Jtag node                        |
+--------------------------+-------+------------------------------------------+
| idCode                   | |int| | IDCODE of Jtag node                      |
+--------------------------+-------+------------------------------------------+
| irLen                    | |int| | IR length Jtag node                      |
+--------------------------+-------+------------------------------------------+
| isTap                    | |int| | flag indicating if node is a JTAG tap    |
+--------------------------+-------+------------------------------------------+
| isMux                    | |int| | flag indicating if node is a mux         |
+--------------------------+-------+------------------------------------------+
| isBranch                 | |int| | flag indicating whether node is branched |
+--------------------------+-------+------------------------------------------+
| isActive                 | |int| | flag indicating whether node is active   |
+--------------------------+-------+------------------------------------------+

"""

from typing import Dict, List, Any, NewType
from chipscopy.tcf import channel
from chipscopy.tcf.services import Service, DoneHWCommand, Token

JtagNode = NewType("JtagNode", Dict[str, Any])

JTAG_SERVICE = "Jtag"


class JtagContext(object):
    """Jtag context class.
    :param props: Properties to initialise this context with. See
    `Properties`_.
    :type props: |dict|
    """

    def __init__(self, service, props):
        self._props = {
            "ID": "",
            "ParentID": "",
            "Name": "",
            "idCode": 0,
            "irLen": 0,
            "isTap": False,
            "isMux": False,
            "isBranch": False,
            "isActive": False,
        }
        self._props.update(props)
        self.service = service

    def __str__(self):
        return "[Jtag Context %s]" % self._props

    @property
    def props(self):
        """Get context properties. See `Properties`_ definitions for property
        names.
        Context properties are read only, clients should not try to modify
        them.
        :returns: A |dict| of context properties.
        """
        return self._props

    def __getattr__(self, item):
        return self._props.get(item)


class JtagListener(object):
    """Jtag event listener is notified when jtag context hierarchy changes,
    and when jtag is modified by Jtag service commands.
    """

    def contextAdded(self, contexts: List[JtagNode]):
        """Called when a new jtag access context(s) is created.
        :param contexts: A list of |JtagContext| properties which have been
        added to jtag space.
        :type contexts: |list|

        """
        pass

    def contextChanged(self, contexts: List[JtagNode]):
        """Called when a jtag access context(s) properties changed.
        :param contexts: A list of |JtagContext| properties which have been
        changed in jtag space.

        :type contexts: |list|

        """
        pass

    def contextRemoved(self, context_ids: List[str]):
        """Called when jtag access context(s) is removed.
        :param context_ids: A list of the IDs of jtag contexts which have
        been removed from jtag space.

        :type context_ids: |list|

        """
        pass


class JtagProxy(Service):
    """TCF Jtag service interface."""

    def __init__(self, channel):
        super(JtagProxy, self).__init__(channel)
        self.listeners = {}

    def getName(self):
        """
        Get this service name.

        :returns: The value of string :const:`NAME`
        """
        return JTAG_SERVICE

    def get_context(self, context_id: str, done: DoneHWCommand) -> Dict[str, Any]:
        """
        Gets context information on a given context id

        :param context_id: Context id of the context
        :param done: Callback with result and any error.
        :return: A dictionary of context parameters
        """
        done = self._makeCallback(done)
        service = self

        class DoneContext(DoneHWCommand):
            def doneHW(self, token, error, args):
                context = None
                if not error and len(args) > 0:
                    context = JtagContext(service, args[0])
                done.doneHW(token, error, context)

        return self.send_command("getContext", (context_id,), DoneContext())

    def get_children(self, context_id: str, done: DoneHWCommand) -> List[str]:
        """
        Gets the children of a given context.  If no context is given then gets top level contexts.

        :param context_id: Context id of desired context
        :param done: Callback with result and any error.
        :return: List of child contexts
        """
        done = self._makeCallback(done)
        if not context_id:
            context_id = ""

        class DoneChildren(DoneHWCommand):
            def doneHW(self, token, error, args):
                context_ids = []
                if not error and len(args) > 0:
                    context_ids = args[0]
                done.doneHW(token, error, context_ids)

        return self.send_command("getChildren", (context_id,), DoneChildren())

    def get_capabilities(self, context_id: str, done: DoneHWCommand) -> Dict[str, Any]:
        """
        Gets the capabilites of a given context.

        :param context_id: Context id of the context
        :param done: Callback with result and any error.
        :return: A dictionary of the capabilities
        """
        done = self._makeCallback(done)
        return self.send_command("getCapabilities", (context_id,), done)

    def set_option(
        self, context_id: str, key: str, value: str or int, done: DoneHWCommand
    ) -> Token:
        """
        Sets an option for a Jtag context

        :param context_id: Context id of the context
        :param key: Name of the option
        :param value: New value of the option
        :param done: Callback with the result and any error
        :return: Token of request
        """
        return self.send_xicom_command("setOption", (context_id, key, value), done)

    def get_option(self, context_id: str, key: str, done: DoneHWCommand) -> Token:
        """
        Retrieves an option value for a specific Jtag context.  The value is returned in the results

        :param context_id: Context id of the context
        :param key: Name of the option
        :param done: Callback with the result and any error
        :return: Token of request
        """
        return self.send_xicom_command("getOption", (context_id, key), done)

    def add_listener(self, listener: JtagListener):
        """Add Jtag service event listener.
        :param listener: Event listener implementation.
        """
        l = ChannelEventListener(self, listener)
        self.channel.addEventListener(self, l)
        self.listeners[listener] = l

    def remove_listener(self, listener: JtagListener):
        """Remove Jtag service event listener.
        :param listener: Event listener implementation.
        """
        l = self.listeners.get(listener)
        if l:
            del self.listeners[listener]
            self.channel.removeEventListener(self, l)


class ChannelEventListener(channel.EventListener):
    def __init__(self, service, listener):
        self.service = service
        self.listener = listener

    def event(self, name, data):
        try:
            args = channel.fromJSONSequence(data)
            if name == "contextAdded":
                assert len(args) == 1
                self.listener.contextAdded(_toContextArray(self.service, args[0]))
            elif name == "contextChanged":
                assert len(args) == 1
                self.listener.contextChanged(_toContextArray(self.service, args[0]))
            elif name == "contextRemoved":
                assert len(args) == 1
                self.listener.contextRemoved(args[0])
            else:
                raise IOError("Memory service: unknown event: " + name)
        except Exception as x:
            import sys

            x.tb = sys.exc_info()[2]
            self.service.channel.terminate(x)


def _toContextArray(svc, o):
    if o is None:
        return None
    ctx = []
    for m in o:
        ctx.append(JtagContext(svc, m))
    return ctx



# %% [markdown]
# <link rel="preconnect" href="https://fonts.gstatic.com">
# <link href="https://fonts.googleapis.com/css2?family=Fira+Code&display=swap" rel="stylesheet">
#
# ### License
#
# <p style="font-family: 'Fira Code', monospace; font-size: 1.2rem">
# Copyright 2021 Xilinx, Inc.<br><br>
# Licensed under the Apache License, Version 2.0 (the "License");<br>
# you may not use this file except in compliance with the License.<br><br>
# You may obtain a copy of the License at <a href="http://www.apache.org/licenses/LICENSE-2.0"?>http://www.apache.org/licenses/LICENSE-2.0</a><br><br>
# Unless required by applicable law or agreed to in writing, software<br>
# distributed under the License is distributed on an "AS IS" BASIS,<br>
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.<br>
# See the License for the specific language governing permissions and<br>
# limitations under the License.<br>
# </p>
#

# %% [markdown]
# # ChipScoPy System Monitor Example
#
#
# <img src="../img/api_overview.png" width="500" align="left">

# %% [markdown]
# ## Description
#
#
# This demo shows how to take and display measurements with the System Monitor.
#
#
# ## Requirements
# - Local or remote Xilinx Versal board, such as a VCK190
# - Xilinx hw_server 2021.2 installed and running
# - Xilinx cs_server 2021.2 installed and running
# - Python 3.8 or greater installed
# - ChipScoPy 2021.2 installed
# - Jupyter notebook support installed - Please do so, using the command `pip install chipscopy[jupyter]`

# %% [markdown]
# ## 1 - Initialization: Imports and File Paths
#
# After this step,
# - Required functions and classes are imported
# - URL paths are set correctly
# - File paths to example files are set correctly

# %%
import os
import time
from chipscopy import get_design_files
from chipscopy import __version__, dm
from chipscopy import create_session, report_versions

# %%
# Make sure to start the hw_server and cs_server prior to running.
# Specify locations of the running hw_server and cs_server below.
# The default is localhost - but can be other locations on the network.
CS_URL = os.getenv("CS_SERVER_URL", "TCP:localhost:3042")
HW_URL = os.getenv("HW_SERVER_URL", "TCP:localhost:3121")

# The get_design_files() function tries to find the PDI and LTX files. In non-standard
# configurations, you can put the path for PROGRAMMING_FILE and PROBES_FILE below.
design_files = get_design_files("vck190/production/chipscopy_ced")

PROGRAMMING_FILE = design_files.programming_file
PROBES_FILE = design_files.probes_file

print(f"HW_URL: {HW_URL}")
print(f"CS_URL: {CS_URL}")
print(f"PROGRAMMING_FILE: {PROGRAMMING_FILE}")
print(f"PROBES_FILE:{PROBES_FILE}")

# %% [markdown]
# ## 2 - Create a session and connect to the hw_server and cs_server
#
# The session is a container that keeps track of devices and debug cores.
# After this step,
# - Session is initialized and connected to server(s)
# - Versions are detected and reported to stdout

# %%
session = create_session(cs_server_url=CS_URL, hw_server_url=HW_URL)
report_versions(session)

# %% [markdown]
# ## 3 - Program the device with the example design
#
# After this step,
# - Device is programmed with the example programming file

# %%
# Typical case - one device on the board - get it.
device = session.devices.filter_by(family="versal").get()
device.program(PROGRAMMING_FILE)

# %% [markdown]
# ## 4 - Discover Debug Cores
#
# Debug core discovery initializes the chipscope server debug cores. This brings debug cores in the chipscope server online.
#
# After this step,
#
# - The cs_server is initialized and ready for use

# %%
device.discover_and_setup_cores(sysmon_scan=True)
print(f"System monitor setup and ready for use.")

# %% [markdown]
# ## 5 - Initialize System Monitor
#
# Get reference to the system monitor and initialize all sensors.

# %%
sysmon = device.sysmon_root.get()

print("Initializing sensors")
active_nodes = sysmon.initialize_sensors()

print("Refresh measurement schedule")
schedule = sysmon.refresh_measurement_schedule()

print("Sensors:")
for sensor in schedule.values():
    print(f"  {sensor}")

print("Done.")


# %% [markdown]
# ## 6 - Register a listener for System Monitor Events
#
# The SysMonNodeListener node_changed() will be called every 1000ms with updated system monitor values.

# %%
class SysMonNodeListener(dm.NodeListener):
    def node_changed(self, node, updated_keys):
        if "device_temp" in node.props:
            print(f"Device Temp: {node.props['device_temp']}")
        for supply_idx, named_sensor in schedule.items():
            supply = f"supply{supply_idx}"
            if supply in node.props:
                print(f"{named_sensor}: {node.props[supply]}")
        print()


node_listener = SysMonNodeListener()
session.chipscope_view.add_node_listener(node_listener)

sysmon.stream_sensor_data(1000)
print("Node listener added.")

# %% [markdown]
# ## 7 - Run measurement for 5 seconds
#
# System Monitor will report results for 5 seconds then exit.

# %%
# Take measurements for 5 seconds then exit.

time_end = time.time() + 5

while time.time() < time_end:
    session.chipscope_view.run_events()
    time.sleep(0.1)

print("Measurement done.")



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum
from dataclasses import dataclass
from typing import NamedTuple, Any, Optional

from chipscopy.api import CoreType, dataclass_fields, filter_props
from chipscopy.api._detail.ltx import Ltx, LtxCore, LtxProbe
from chipscopy.client.axis_ila_core_client import (
    AxisIlaCoreClient as TCF_AxisIlaCoreClient,
    ILAPortType as TCF_ILAPortType,
)


class ILAMatchUnitType(enum.Enum):
    """
    Probe comparator types. Only LARGE is supported.

    ===========  =====================================================================
    Enum Value   Description
    ===========  =====================================================================
    SMALL        Not supported.
    MEDIUM       Not supported.
    LARGE        Supports operators: ['==', '!=', '>', '<', '>=', '<=', '||']
                  and bit values: '01xXrRfFnNbBtTlLsS\_'
    ===========  =====================================================================

    """

    SMALL = 0
    MEDIUM = 1
    LARGE = 2


class ILAProbeRadix(enum.Enum):
    """
    Display radix for probe data values.

    =====================  ================================
    Enum Value             Description
    =====================  ================================
    BIN                    Not Supported.
    ENUM                   Enumeration
    HEX                    Default.
    SIGNED                 Not supported.
    UNSIGNED               Not supported.
    =====================  ================================

    """

    BIN = 0
    ENUM = 1
    HEX = 2
    SIGNED = 3
    UNSIGNED = 4


@dataclass(frozen=True)
class ILAPort:
    """Probe Port on the ILA Core"""

    bit_width: int
    """Number of port bits."""
    data_bit_index: int
    """Start bit index in waveform data."""
    index: int
    """Port index"""
    is_trigger: bool
    """Port is used for triggering."""
    is_data: bool
    """Port is used for capturing data."""
    match_unit_count: int
    """Number of compare match units, available to the port."""
    match_unit_type: ILAMatchUnitType
    """Type of match unit."""


ILA_PORT_MEMBERS = dataclass_fields(ILAPort)


@dataclass(frozen=True)
class ILAProbe:
    """Logical probe. Probes may share the same probe port."""

    name: str
    """Name of probe."""
    bit_width: int
    """Number of probe bits."""
    is_data: bool
    """Probe is used to capture waveform data."""
    is_trigger: bool
    """Probe can be used in trigger settings."""
    map: str
    """How probe is mapped to ports."""
    mu_count: int
    """Max number of compare match units, available to the probe."""
    is_bus: bool = False
    """True for bus probes"""
    bus_left_index: Optional[int] = None
    """Bus left index. E.g. 5 in probe ``counter[5:0]``"""
    bus_right_index: Optional[int] = None
    """Bus right index. E.g. 0 in probe ``counter[5:0]``"""


ILA_PROBE_MEMBERS = dataclass_fields(ILAProbe)


ILA_MATCH_OPERATORS = ["==", "!=", ">", "<", ">=", "<=", "||"]
"""
Operators for probe compare values.

========= =================================================================
Operator  Description
========= =================================================================
==        Equal
!=        Not equal
<         Less than
<=        Equal or less than
>         Greater than
>=        Equal or Greater than
||        Reduction OR
========= =================================================================

"""


ILA_MATCH_BIT_VALUES = "01xXrRfFnNbBtTlLsS_"
"""
Binary bit values for probe compare values.

========= =================================================================
Bit Value Description
========= =================================================================
_         Underscore separator for readability.
X         Don't care bit matches any bit-value.
0         Zero
1         One
F         Falling. Transition 1 -> 0
R         Rising. Transition 0 -> 1
L         Laying. Opposite to R.
S         Staying. Opposite to F.
B         Either Falling or Rising.
N         No change. Opposite to B.
========= =================================================================

"""


@dataclass
class ILAProbeValues:
    """
    A compare value consists of a pair: <operator> and <value>
    <value> may be of type int or binary/hex string of the correct bit_width.
    Hex values start with a '0x' prefix.

    An empty list is a don't-care value.




        Example: Test for LSB is '0' on a 4-bit probe.
           ['==',   'XXX0']

        Example: Range check, assuming the global Trigger Condition is ILATriggerCondition.AND:
           ['>=',   '0x3',   '<=',   '0x10']

        Example: Range check, assuming the global Trigger Condition is ILATriggerCondition.AND:
           ['>=',   3,   '<=',   10]

        Example: Test equal to any of 3 numbers, assuming the global Trigger Condition is ILATriggerCondition.OR:
           ['==',   '0011',   '==',   '0111',   '==',   '1111']

        Example: 4-bit dont-care value.
           ['==',   'XXXX']

        Example: Dont-care value, written as an empty list.
           []

    """

    bit_width: int
    """Number of probe bits."""
    trigger_value: []
    """Trigger compare values, in a list."""
    capture_value: []
    """Basic capture compare value. A two item list with [<operator>, <value>]"""
    display_radix: ILAProbeRadix = ILAProbeRadix.HEX
    """Display radix, when exporting waveform data. Default is ILAProbeRadix.HEX"""
    enum_def: Optional[enum.EnumMeta] = None
    """Enum class defining {name:int} enum values, for this probe."""

    def default_value(self):
        return ["==", "x" * self.bit_width]


class ILABitRange(NamedTuple):
    index: int
    """Bit index."""
    length: int
    """Bit length."""


#
# ports functions
#
def verify_ports(ltx: Ltx, uuid: str, ports: [ILAPort]) -> None:
    def get_width(port: ILAPort, trigger: bool, data: bool) -> int:
        return port.bit_width if port.is_trigger == trigger and port.is_data == data else 0

    widths_by_type = {
        "DATA": [get_width(port, trigger=False, data=True) for port in ports],
        "DATA_TRIGGER": [get_width(port, trigger=True, data=True) for port in ports],
        "TRIGGER": [get_width(port, trigger=True, data=False) for port in ports],
    }
    ltx.verify_port_width_by_type(CoreType.AXIS_ILA, uuid, widths_by_type)


def ports_from_tcf_props(props) -> [ILAPort]:
    ports = []
    for p_index, p in enumerate(props["ports"]):
        port_info = filter_props(p, ILA_PORT_MEMBERS)
        port_info["index"] = p_index
        port_info["match_unit_count"] = len(p["mus"])
        port_type = p["port_type"]
        port_info["is_data"] = port_type & TCF_ILAPortType.IS_DATA != 0
        port_info["is_trigger"] = port_type & TCF_ILAPortType.IS_TRIGGER != 0
        if port_type & TCF_ILAPortType.MU_EDGE_EQ_REL:
            port_info["match_unit_type"] = ILAMatchUnitType.LARGE
        elif port_type & TCF_ILAPortType.MU_EDGE_EQ_EDGE:
            port_info["match_unit_type"] = ILAMatchUnitType.MEDIUM
        else:
            port_info["match_unit_type"] = ILAMatchUnitType.SMALL
        ports.append(ILAPort(**port_info))

    return ports


#
# probes functions
#
def verify_probe_value(value_list: [], is_trigger: bool, probe: ILAProbe, enum_def=None):
    """
    int values, enums, binary string and hex strings are supported for now.
    More verification is done by the cs_server.

    Args:
        enum_def ():
    """

    def is_valid_enum_str(number) -> bool:
        try:
            enum = enum_def[number]
        except KeyError:
            return False
        verify_enum_value(enum)
        return True

    def get_hex_char_bit_len(ch: str) -> int:
        if ch in "x01":
            return 1
        elif ch in "23":
            return 2
        elif ch in "4567":
            return 3
        else:
            return 4

    def verify_int_bit_length(number):
        val_len = int.bit_length(number)
        if val_len > probe.bit_width:
            raise ValueError(
                f'Probe "{probe.name}" has bit width:{probe.bit_width} and cannot be assigned'
                f' value "{number}" which has bit width:{val_len}.'
            )

    def verify_enum_value(enum: enum.Enum):
        if not isinstance(enum, enum_def):
            raise TypeError(f'Probe "{probe.name}" cannot be assigned enum value {enum}.')
        verify_int_bit_length(enum.value)

    def verify_one_hex_value(op: str, hex_number: str):
        valid_hex_chars = "x0123456789abcdef_"
        value_no_underscore = "".join([ch for ch in hex_number[2:].lower() if ch != "_"])
        hex_val_len = len(value_no_underscore)
        if hex_val_len == 0:
            raise ValueError(
                f'Probe "{probe.name}" has cannot be assigned invalid value "{hex_number}".'
            )
        if not all([ch in valid_hex_chars for ch in value_no_underscore]):
            raise ValueError(
                f'Probe "{probe.name}" value "{hex_number}" has invalid character(s).'
                f' Valid characters are {valid_hex_chars}".'
            )
        required_hex_ch_len = (probe.bit_width + 3) // 4
        if hex_val_len != required_hex_ch_len:
            raise ValueError(
                f'Probe "{probe.name}" has bit width:{probe.bit_width} requiring {required_hex_ch_len} hex character(s)'
                f' but value "{hex_number}" has only {hex_val_len} hex character(s).'
            )
        bin_val_len = (hex_val_len - 1) * 4 + get_hex_char_bit_len(value_no_underscore[0])
        if bin_val_len > probe.bit_width:
            # Too many bits in first hex char.
            raise ValueError(
                f'Probe "{probe.name}" has bit width:{probe.bit_width} and cannot be assigned'
                f' value "{hex_number}" which has bit width:{bin_val_len}.'
            )
        if (op not in ["==", "!="]) and ("x" in value_no_underscore):
            raise ValueError(
                f'Probe "{probe.name}" is assigned value "{op} {hex_number}". '
                f"""The operator cannot be used when the value has a don't care "x"."""
            )

    def verify_value_pair(op: str, number):
        if op not in ILA_MATCH_OPERATORS:
            raise ValueError(
                f'Probe "{probe.name}" operator "{op}" is not one of the allowed operators: '
                f"{ILA_MATCH_OPERATORS}."
            )

        if isinstance(number, int):
            verify_int_bit_length(number)
        elif enum_def and isinstance(number, enum.Enum):
            verify_enum_value(number)
        elif isinstance(number, str) and (number.startswith("0x") or number.startswith("0X")):
            verify_one_hex_value(op, number)
        elif enum_def and is_valid_enum_str(number):
            # is_valid_enum_str() will did the checking if it is an enum string for this probe.
            pass
        elif isinstance(number, str):
            # binary value
            if not all([ch in ILA_MATCH_BIT_VALUES for ch in number]):
                raise TypeError(
                    f'Probe "{probe.name}" value "{number}" has invalid character(s).'
                    f' Valid characters are "{ILA_MATCH_BIT_VALUES}".'
                )

            val_len = len(number) - number.count("_")
            if val_len != probe.bit_width:
                raise ValueError(
                    f'Probe "{probe.name}" has bit width:{probe.bit_width} and cannot be assigned'
                    f' value "{number}" which has bit width:{val_len}.'
                )
            if (op not in ["==", "!="]) and not all([ch in "01_" for ch in number]):
                raise TypeError(
                    f'Probe "{probe.name}" is assigned value "{op} {number}". '
                    f'The operator "{op}" can only be used with "0" and "1" bit values.'
                )

    # Empty list are allowed. Used for DONT_CARE values.
    if not isinstance(value_list, list):
        raise TypeError(
            f'Probe "{probe.name}" trigger/capture set value {value_list} is required to be a list.'
        )
    if len(value_list) > 2 and not is_trigger:
        raise ValueError(
            f'Probe "{probe.name}" capture set value {value_list} list length cannot be more than 2.'
        )
    if len(value_list) % 2 != 0:
        raise TypeError(
            f'Probe "{probe.name}" trigger/capture set value {value_list} is required to be a list'
            " with even number of items. Odd items are operators and even items are numbers."
        )
    for idx in range(0, len(value_list), 2):
        verify_value_pair(value_list[idx], value_list[idx + 1])


def create_probes_from_ports_and_ltx(
    tcf_ila: TCF_AxisIlaCoreClient, ports: [ILAPort], ltx: Ltx, uuid: str
) -> ({}, {}, str, {}):
    """ Returns:  probes, map_to_port_seqs, cell_name, probe enum defs"""

    def process_one_tcf_probe(attrs: {str, Any}) -> {}:
        res = {
            key: value
            for key, value in attrs.items()
            if key in {"bit_width", "mu_count", "map", "name"}
        }
        port_type = TCF_ILAPortType(attrs.get("port_type", 0))
        res["is_data"] = TCF_ILAPortType.IS_DATA in port_type
        res["is_trigger"] = TCF_ILAPortType.IS_TRIGGER in port_type
        return res

    def process_one_tcf_probe_mapping(attrs: {str, Any}) -> [ILABitRange]:
        def create_bit_range(map_seq: ()) -> ILABitRange:
            port_idx, high_idx, low_idx = map_seq
            start = ports[port_idx].data_bit_index + low_idx
            length = abs(high_idx - low_idx) + 1
            return ILABitRange(start, length)

        tcf_mapping = attrs.get("__map_seqs", [])
        mapping = [create_bit_range(map_seq) for map_seq in tcf_mapping]
        return mapping

    def ltx_probes_to_tcf_pdefs(ltx: LtxCore) -> [{str, str}]:
        return [
            {
                "name": probe.name,
                "net_name": probe.name,
                "map": f"port{probe.port_index}[{probe.port_msb_index}:{probe.port_lsb_index}]",
            }
            for probe in ltx.probes
            if probe.probe_type != "DATA"
        ]

    def add_probe_bus_info(p_dicts: {}, ltx_probes: [LtxProbe]) -> None:
        for ltx_p in ltx_probes:
            if ltx_p.is_bus and ltx_p.name in p_dicts:
                p_dicts[ltx_p.name].update(
                    {
                        "is_bus": ltx_p.is_bus,
                        "bus_left_index": ltx_p.bus_left_index,
                        "bus_right_index": ltx_p.bus_right_index,
                    }
                )

    def get_probe_enum_defs(p_dicts: {}, ltx_probes: [LtxProbe]) -> {}:
        res = {}
        for ltx_p in ltx_probes:
            if ltx_p.enum_def and ltx_p.name in p_dicts:
                res[ltx_p.name] = ltx_p.enum_def
        return res

    cell_name = ""
    tcf_ila.undefine_probe([])
    ltx_core = ltx.get_core(CoreType.AXIS_ILA, uuid) if ltx else None
    if ltx and not ltx_core:
        raise IndexError(f"Unable to find ILA core with uuid:{uuid} in LTX file.")
    if ltx_core:
        cell_name = ltx_core.cell_name
        probe_defs = ltx_probes_to_tcf_pdefs(ltx_core)
        tcf_ila.define_probe(probe_defs)
    else:
        tcf_ila.define_port_probes({"force": True})

    probe_attrs = ["name", "bit_width", "mu_count", "port_type", "map", "__map_seqs"]
    tcf_probes = tcf_ila.get_probe([], probe_attrs)
    probe_dicts = {p_name: process_one_tcf_probe(attrs) for p_name, attrs in tcf_probes.items()}
    # Add bus information
    add_probe_bus_info(probe_dicts, ltx_core.probes)
    enum_defs = get_probe_enum_defs(probe_dicts, ltx_core.probes)
    map_to_port_seqs = {
        attrs["map"]: process_one_tcf_probe_mapping(attrs) for p_name, attrs in tcf_probes.items()
    }
    probes = {p_name: ILAProbe(**probe) for p_name, probe in probe_dicts.items()}
    return probes, map_to_port_seqs, cell_name, enum_defs


def verify_probe_enum_def(probe: ILAProbe, enum_def: enum.EnumMeta):
    if not isinstance(enum_def, enum.EnumMeta):
        raise TypeError(
            f"enum definition need to be an Enum class. Type '{type(enum_def)}' was passed in."
        )
    for ei in enum_def:
        if type(ei.value) is not int:
            raise TypeError(f'enum {ei} value must be of type "int" for probe "{probe.name}".')
        if ei.value.bit_length() > probe.bit_width:
            raise ValueError(
                f"Enum value {ei} has bit length:{ei.value.bit_length()}, "
                f'which is more than probe "{probe.name}" bit length:{probe.bit_width}.'
            )



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. |state| replace::   :class:`ILAState`
.. |cc|    replace::   :class:`ILACaptureCondition`
.. |cmod|  replace::   :class:`ILACaptureMode`
.. |tc|    replace::   :class:`ILATriggerCondition`
.. |tmod|  replace::   :class:`ILATriggerMode`
.. |tout|  replace::   :class:`ILATrigOutMode`
.. |prtd|  replace::   :class:`ILAPort`
.. |prtt|  replace::   :class:`ILAPortType`

ILA Service Overview
---------------------

The ILA TCF Service handles one ILA core.

*Data Model*

The ILA features are represented by "properties". The properties are divided into groups:

* `Core Info Properties`_ - Core version and type information.
* `Static Info Properties`_ - Core capabilities and dimensions.
* `Status Properties`_ - Current core dynamic state.
* `Control Properties`_ - Control of the core.
* `Data Properties`_ - Uploaded waveform.

*Methods*

The Service methods are of the following types:

* `Property Methods`_ - Set/get/reset/report properties.
* `Probe Methods`_ - Create/delete probes. Set/get probe match values.
* `Core Communication Methods`_ - Read/write properties from/to core. Arm the core and upload waveform.

Service Name
------------
.. autodata:: NAME

Properties
----------
Property Groups
^^^^^^^^^^^^^^^
Properties are divided into groups. Properties within a group are often have
their values 'set' or 'get' at the same time.

+--------------------------+-------+------------------------------------------+
| Name                     | Type  | Description                              |
+==========================+=======+==========================================+
| PROP_GROUP_CORE_INFO     | |str| | Core version and type information.       |
+--------------------------+-------+------------------------------------------+
| PROP_GROUP_STATIC_INFO   | |str| | Read-only properties giving              |
|                          |       | capabilities and dimensions.             |
+--------------------------+-------+------------------------------------------+
| PROP_GROUP_STATUS        | |str| | Read-only properties showing current     |
|                          |       | dynamic state.                           |
+--------------------------+-------+------------------------------------------+
| PROP_GROUP_CONTROL       | |str| | Write/Read properties to control core.   |
+--------------------------+-------+------------------------------------------+
| PROP_GROUP_DATA          | |str| | Read-only properties holding waveform.   |
+--------------------------+-------+------------------------------------------+

Core Info Properties
^^^^^^^^^^^^^^^^^^^^^^^^
*Core Info* are read-only properties with core type and version information.

+--------------------------+--------+--------------------------------------------+
| Name                     | Type   | Description                                |
+==========================+========+============================================+
| PROP_CORE_TYPE           | |int|  | Core type number.                          |
+--------------------------+--------+--------------------------------------------+
| PROP_DRV_VER             | |int|  | Software core interface version.           |
+--------------------------+--------+--------------------------------------------+
| PROP_CORE_MAJOR_VER      | |int|  | Core major version.                        |
+--------------------------+--------+--------------------------------------------+
| PROP_CORE_MINOR_VER      | |int|  | Core minor version.                        |
+--------------------------+--------+--------------------------------------------+
| PROP_TOOL_MAJOR_VER      | |int|  | Major version of tool, which created core. |
+--------------------------+--------+--------------------------------------------+
| PROP_TOOL_MAJOR_VER      | |int|  | Minor version of tool, which created core. |
+--------------------------+--------+--------------------------------------------+
| PROP_UUID                | |str|  | Hex value, unique core id within design.   |
+--------------------------+--------+--------------------------------------------+

Static Info Properties
^^^^^^^^^^^^^^^^^^^^^^^^^^
*Static Info* are read-only properties giving the dimensions and feature set of
the core.

+---------------------------+--------+-----------------------------------------+
| Name                      | Type   | Description                             |
+===========================+========+=========================================+
| PROP_DATA_DEPTH           | |int|  | Max sample depth of waveform data.      |
+---------------------------+--------+-----------------------------------------+
| PROP_DATA_WIDTH           | |int|  | Bit width of waveform data.             |
+---------------------------+--------+-----------------------------------------+
| PROP_HAS_ADVANCED_TRIGGER | |bool| | True if core has advanced trigger.      |
+---------------------------+--------+-----------------------------------------+
| PROP_HAS_CAPTURE_CONTROL  | |bool| | True if core has capture control.       |
+---------------------------+--------+-----------------------------------------+
| PROP_HAS_TRIG_IN          | |bool| | True if core has trig_in port.          |
+---------------------------+--------+-----------------------------------------+
| PROP_HAS_TRIG_OUT         | |bool| | True if core has trig_out port.         |
+---------------------------+--------+-----------------------------------------+
| PROP_MATCH_UNIT_COUNT     | |bool| | Number of port match units.             |
+---------------------------+--------+-----------------------------------------+
| PROP_PORT_COUNT           | |int|  | Number of ports.                        |
+---------------------------+--------+-----------------------------------------+
| PROP_PORTS                |[|prtd|]| Ordered list of port definitions.       |
+---------------------------+--------+-----------------------------------------+

Status Properties
^^^^^^^^^^^^^^^^^^^
*Status* read-only properties showing the dynamic status of the core.

+--------------------------+--------+-------------------------------------------+
| Name                     | Type   | Description                               |
+==========================+========+===========================================+
| PROP_CAPTURE_STATE       | |state|| State of capture controller.              |
+--------------------------+--------+-------------------------------------------+
| PROP_IS_ARMED            | |bool| | Capture control is armed.                 |
+--------------------------+--------+-------------------------------------------+
| PROP_IS_FULL             | |bool| | True if waveform data buffer is full.     |
+--------------------------+--------+-------------------------------------------+
| PROP_SAMPLES_CAPTURED    | |int|  | Current window captured samples count.    |
+--------------------------+--------+-------------------------------------------+
| PROP_WINDOWS_CAPTURED    | |bool| | Number of fully captured windows.         |
+--------------------------+--------+-------------------------------------------+

Control Properties
^^^^^^^^^^^^^^^^^^^^
*Control* read/write properties are used to control the core.

+--------------------------+--------+-----------------------------------------+
| Name                     | Type   | Description                             |
+==========================+========+=========================================+
| PROP_ARM                 | |bool| | Set to True, to arm the core.           |
+--------------------------+--------+-----------------------------------------+
| PROP_CAPTURE_CONDITION   | |cc|   | Capture condition: AND, OR, NAND, NOR   |
+--------------------------+--------+-----------------------------------------+
| PROP_CAPTURE_MODE        | |cmod| | Capture mode: always or basic.          |
+--------------------------+--------+-----------------------------------------+
| PROP_TRIG_OUT_MODE       | |tout| | Trig out signal choices.                |
+--------------------------+--------+-----------------------------------------+
| PROP_TRIGGER_CONDITION   | |tc|   | Trigger Condition: AND, OR, NAND, NOR,  |
|                          |        |    IMMEDIATELY.                         |
+--------------------------+--------+-----------------------------------------+
| PROP_TRIGGER_MODE        | |tmod| | Trigger mode. What to trigger on.       |
+--------------------------+--------+-----------------------------------------+
| PROP_TRIGGER_POSITION    | |int|  | Trigger position within window.         |
|                          |        | 0 <= trigger position < window size     |
+--------------------------+--------+-----------------------------------------+
| PROP_WINDOW_COUNT        | |int|  | Number of windows.                      |
+--------------------------+--------+-----------------------------------------+
| PROP_WINDOW_SIZE         | |int|  | Sample count per window. The value must |
|                          |        | be a power-of-two.                      |
+--------------------------+--------+-----------------------------------------+

Data Properties
^^^^^^^^^^^^^^^^^
The *Data* read-only properties represent the uploaded waveform.

+--------------------------+-----------------+--------------------------------+
| Name                     | Type            | Description                    |
+==========================+=================+================================+
| trace_data               | |bytearray|     | Waveform bit values.           |
+--------------------------+-----------------+--------------------------------+
| trace_width              | |int|           | Data sample bit width.         |
+--------------------------+-----------------+--------------------------------+
| trace_sample_count       | |int|           | Captured sample count.         |
+--------------------------+-----------------+--------------------------------+
| trace_trigger_position   | [|int|]         | List of trigger positions.     |
|                          |                 | One value, per window.         |
+--------------------------+-----------------+--------------------------------+
| trace_window_size        | |int|           | Sample count per window.       |
+--------------------------+-----------------+--------------------------------+

The samples in the trace_data are aligned on byte boundaries.

This formula can be used to read a bit from the trace_data:
::

    bytes_per_sample = len(trace_data) // trace_sample_count

    def get_bit_value(trace_data: bytearray, bytes_per_sample: int, sample_index: int, data_bit_index: int) -> bool:
        byte_value = trace_data[sample_index * bytes_per_sample + data_bit_index // 8]
        mask = 1 << (data_bit_index & 0x7)
        return (byte_value & mask) != 0

Property Value Types
--------------------

PropertyPermission Enum Values
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
.. autoclass:: PropertyPermission
    :members:
    :undoc-members:

Capture State Enum Values
^^^^^^^^^^^^^^^^^^^^^^^^^
.. autoclass:: chipscopy.client.axis_ila_core_client.ILAState
    :members:
    :undoc-members:

Trigger Mode Enum Values
^^^^^^^^^^^^^^^^^^^^^^^^
.. autoclass:: chipscopy.client.axis_ila_core_client.ILATriggerMode
    :members:
    :undoc-members:

Trigger Condition Enum Values
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
.. autoclass:: chipscopy.client.axis_ila_core_client.ILATriggerCondition
    :members:
    :undoc-members:

Capture Mode Enum Values
^^^^^^^^^^^^^^^^^^^^^^^^
.. autoclass:: chipscopy.client.axis_ila_core_client.ILACaptureMode
    :members:
    :undoc-members:

Capture Condition Enum Values
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
.. autoclass:: chipscopy.client.axis_ila_core_client.ILACaptureCondition
    :members:
    :undoc-members:

Trig Out Mode Enum Values
^^^^^^^^^^^^^^^^^^^^^^^^^
.. autoclass:: chipscopy.client.axis_ila_core_client.ILATrigOutMode
    :members:
    :undoc-members:




Param Types
-----------

PropertyValues
^^^^^^^^^^^^^^
.. autodata:: PropertyValues

Dict to hold property values.

::

    PropertyValues = {str, Any}


=============== ================
Key             Value
=============== ================
property name   property value
=============== ================

.. _ILAPort:

ILAPort
^^^^^^^^^^^^
.. autodata:: ILAPort

Dict type to hold port attributes.

::

    ILAPort = Dict[str, Any]


============== ======= ========================================================
Key            Type    Value
============== ======= ========================================================
data_bit_index int     Data sample start bit index. '-1' for trigger-only port.
index          int     Port index.
mu_count       int     Port match unit count.
mus            [|int|] List of match unit indices.
port_type      |prtt|  Capabilities of the port.
bit_width      int     Port bit width.
============== ======= ========================================================

.. _ILAPortType:

ILAPortType Enum Constants
^^^^^^^^^^^^^^^^^^^^^^^^^^
.. autoclass:: chipscopy.client.axis_ila_core_client.ILAPortType
    :members:
    :undoc-members:


.. _ILAProbeAttrs:

ILAProbeAttrs
^^^^^^^^^^^^^
.. autodata:: ILAProbeAttrs

Dict type to hold probe attributes.
::

    ILAProbeAttrs = Dict[str, Any]

================= =========== ==============================================
Key               Value Type  Description
================= =========== ==============================================
bit_width         int         Probe bit width.
capture_value     |str|       One match value string, for capture control.
name              |str|       Probe name.
net_name          |str|       Netlist name
map               |str|       Definition string for probe mapping.
mu_count          |int|       Number of match units for trigger.
port_type         |prtt|      Port type associated with trigger probe.
trigger_value     [|str|]     mu_count of match value strings.
================= =========== ==============================================

.. _ila_prop_match_value_table:

Match values consist of an operator followed by one match character per bit.
::

    Example:
        { 'capture_value':  '==0001_XXXX',
          'trigger_value': ['<=0011_1111', '>0000_0111'] }

========= =================================================================
Operator  Description
========= =================================================================
==        Equal
!=        Not equal
<         Less than
<=        Equal or less than
>         Greater than
>=        Equal or Greater than
||        Reduction OR
========= =================================================================



========= =================================================================
Bit Value Description
========= =================================================================
_         Underscore separator for readability.
X         Don't care bit matches any bit-value.
0         Zero
1         One
F         Falling. Transition 1 -> 0
R         Rising. Transition 0 -> 1
L         Laying. Opposite to R.
S         Staying. Opposite to F.
B         Either Falling or Rising.
N         No change. Opposite to B.
========= =================================================================



General Methods
---------------

initialize
^^^^^^^^^^
.. automethod:: AxisIlaCoreClient.initialize

terminate
^^^^^^^^^
.. automethod:: AxisIlaCoreClient.terminate

Property Methods
----------------

get_property
^^^^^^^^^^^^
.. automethod:: AxisIlaCoreClient.get_property

get_property_group
^^^^^^^^^^^^^^^^^^
.. automethod:: AxisIlaCoreClient.get_property_group

report_property
^^^^^^^^^^^^^^^
.. automethod:: AxisIlaCoreClient.report_property

report_property_group
^^^^^^^^^^^^^^^^^^^^^
.. automethod:: AxisIlaCoreClient.report_property_group

reset_property
^^^^^^^^^^^^^^
.. automethod:: AxisIlaCoreClient.reset_property

reset_property_group
^^^^^^^^^^^^^^^^^^^^
.. automethod:: AxisIlaCoreClient.reset_property_group

set_property
^^^^^^^^^^^^
.. automethod:: AxisIlaCoreClient.set_property

set_property_group
^^^^^^^^^^^^^^^^^^
.. automethod:: AxisIlaCoreClient.set_property_group


Probe Methods
-------------

define_probe
^^^^^^^^^^^^
.. automethod:: AxisIlaCoreClient.define_probe

define_port_probes
^^^^^^^^^^^^^^^^^^
.. automethod:: AxisIlaCoreClient.define_port_probes

undefine_probe
^^^^^^^^^^^^^^
.. automethod:: AxisIlaCoreClient.undefine_probe

get_probe
^^^^^^^^^
.. automethod:: AxisIlaCoreClient.get_probe

set_probe
^^^^^^^^^
.. automethod:: AxisIlaCoreClient.set_probe

Core Communication Methods
--------------------------

arm
^^^
.. automethod:: AxisIlaCoreClient.arm

commit_property_group
^^^^^^^^^^^^^^^^^^^^^
.. automethod:: AxisIlaCoreClient.commit_property_group

refresh_property_group
^^^^^^^^^^^^^^^^^^^^^^
.. automethod:: AxisIlaCoreClient.refresh_property_group


upload
^^^^^^
.. automethod:: AxisIlaCoreClient.upload

"""

import enum
import time
from typing import List, Dict, Any

from chipscopy import dm
from chipscopy.client.core_property_client import CorePropertyClient
from chipscopy.tcf.services import DoneHWCommand
from chipscopy.tcf import protocol

# Service name.
NAME = "AxisILA"
"""
ILA service name.
"""


PROP_GROUP_CORE_INFO = "core_info"
PROP_GROUP_STATIC_INFO = "static_info"
PROP_GROUP_STATUS = "status"
PROP_GROUP_CONTROL = "control"
PROP_GROUP_DATA = "data"
PROP_CORE_TYPE = "core_type"
PROP_DRV_VER = "drv_ver"
PROP_CORE_MAJOR_VER = "core_major_ver"
PROP_CORE_MINOR_VER = "core_minor_ver"
PROP_TOOL_MAJOR_VER = "tool_major_ver"
PROP_TOOL_MINOR_VER = "tool_minor_ver"
PROP_UUID = "uuid"
PROP_DATA_DEPTH = "data_depth"
PROP_DATA_WIDTH = "data_width"
PROP_HAS_ADVANCED_TRIGGER = "has_advanced_trigger"
PROP_HAS_CAPTURE_CONTROL = "has_capture_control"
PROP_HAS_TRIG_IN = "has_trig_in"
PROP_HAS_TRIG_OUT = "has_trig_out"
PROP_MATCH_UNIT_COUNT = "match_unit_count"
PROP_PORT_COUNT = "port_count"
PROP_PORTS = "ports"
PROP_CAPTURE_STATE = "capture_state"
PROP_IS_ARMED = "is_armed"
PROP_IS_FULL = "is_full"
PROP_IS_TAS = "is_tas"
PROP_SAMPLE_COUNT_STATUS = "samples_captured"
PROP_WINDOW_COUNT_STATUS = "windows_captured"
PROP_ARM = "arm_ila"
PROP_CAPTURE_CONDITION = "capture_condition"
PROP_CAPTURE_MODE = "capture_mode"
PROP_TRIG_OUT_MODE = "trig_out_mode"
PROP_TRIGGER_CONDITION = "trigger_condition"
PROP_TRIGGER_MODE = "trigger_mode"
PROP_TRIGGER_POSITION = "trigger_position"
PROP_WINDOW_COUNT = "window_count"
PROP_WINDOW_SIZE = "window_size"
PROP_TRACE_DATA = "trace_data"
PROP_TRACE_WIDTH = "trace_width"
PROP_TRACE_SAMPLE_COUNT = "trace_sample_count"
PROP_TRACE_TRIGGER_POSITION = "trace_trigger_position"
PROP_TRACE_WINDOW_SIZE = "trace_window_size"


class ILAPortType(enum.IntFlag):
    """

        Capabilities of a port.

        MU_EQ supports eq, neq match operators.

        MU_EDGE supports edge transition match bit values: 'R', 'F', 'B', 'N'

        MU_REL supports relational match operators: lt, gt, lte, gte.

        When passing enum values to/from API functions, use int values.

    ::

        IS_DATA    = 0x01
        IS_TRIGGER = 0x02
        MU_EDGE    = 0x04
        MU_EQ      = 0x08
        MU_REL     = 0x10

        IS_TRIGGER_DATA = IS_TRIGGER | IS_DATA
        MU_EDGE_EQ_REL = MU_EDGE | MU_EQ | MU_REL


    """

    IS_DATA = 0x01
    IS_TRIGGER = 0x02
    MU_EDGE = 0x04
    MU_EQ = 0x08
    MU_REL = 0x10
    IS_TRIGGER_DATA = IS_TRIGGER | IS_DATA
    MU_EDGE_EQ_REL = MU_EDGE | MU_EQ | MU_REL


ILAPort = Dict[str, Any]
ILAProbeAttrs = Dict[str, Any]


class ILAState(enum.Enum):
    """
        *Trigger states:*

        State transitions, in case of one window.

        IDLE -> PRE_TRIGGER -> TRIGGER -> POST_TRIGGER -> IDLE

        When passing enum values to/from API functions, use string values.

    ::

        IDLE
        PRE_TRIGGER
        TRIGGER
        POST_TRIGGER
    """

    IDLE = 0
    PRE_TRIGGER = 1
    TRIGGER = 2
    POST_TRIGGER = 3


class ILATrigOutMode(enum.Enum):
    """
        Trig_out signal control choices.

        When passing enum values to/from API functions, use string values.

    ::

        DISABLED
        TRIGGER_ONLY
        TRIG_IN_ONLY
        TRIGGER_OR_TRIG_IN
    """

    DISABLED = 0
    TRIGGER_ONLY = 1
    TRIG_IN_ONLY = 2
    TRIGGER_OR_TRIG_IN = 3


class ILATriggerCondition(enum.Enum):
    """
        Trigger condition combining Match Units.

        When passing enum values to/from API functions, use string values.

    ::

        OR
        AND
        NOR
        NAND
        IMMEDIATELY
    """

    OR = 0
    AND = 1
    NOR = 2
    NAND = 3
    IMMEDIATELY = 4


class ILATriggerMode(enum.Enum):
    """
        Choice between basic mode using probe match values, advanced sequential trigger,
        trig_in or a valid combination of them.

        When passing enum values to/from API functions, use string values.

    ::

        BASIC_ONLY
        ADVANCED_ONLY
        BASIC_OR_TRIG_IN
        ADVANCED_OR_TRIG_IN
        TRIG_IN_ONLY
    """

    BASIC_ONLY = 0
    ADVANCED_ONLY = 1
    BASIC_OR_TRIG_IN = 2
    ADVANCED_OR_TRIG_IN = 3
    TRIG_IN_ONLY = 4


class ILACaptureCondition(enum.Enum):
    """
        OR, AND, NAND, NOR are choices to combine probe match values.

        When passing enum values to/from API functions, use string values.

    ::

        OR
        AND
        NOR
        NAND
    """

    OR = 0
    AND = 1
    NOR = 2
    NAND = 3


class ILACaptureMode(enum.Enum):
    """
    BASIC mode, will use capture condition to decide which samples to keep.

    When passing enum values to/from API functions, use string values.
    """

    ALWAYS = 0
    BASIC = 1


class AxisIlaCoreClient(CorePropertyClient):
    """TCF AXIS-ILA Client"""

    @staticmethod
    def is_compatible(node: dm.Node) -> bool:
        return node.type and node.type == "ila"

    def get_service_proxy(self):
        return self.manager.channel.getRemoteService(NAME)

    #
    #  General methods
    #
    def initialize(self, done: DoneHWCommand) -> None:
        """
        Initialize data model for the ILA core.

        :param done: Callback with result and any error.
        """
        service, done_cb = self.make_done(done)
        token = service.initialize(self.ctx, done_cb)
        return self.add_pending(token)

    def terminate(self, done: DoneHWCommand) -> None:
        """
        Remove data model for the ILA core.

        :param done: Callback with result and any error.
        """
        service, done_cb = self.make_done(done)
        token = service.terminate(self.ctx, done_cb)
        return self.add_pending(token)

    #
    # Probe Methods
    #
    def define_probe(self, probe_defs: List[Dict[str, Any]], done: DoneHWCommand) -> None:
        """
        Define ILA probe(s).

        A probe has the following attributes, stored in a :ref:`ILAProbeAttrs` dict:

        ============== ========== ====================================
        Attribute      Value Type Description
        ============== ========== ====================================
        name           |str|      Probe name.
        net_name       |str|      Netlist name
        map            |str|      Definition string for probe mapping.
        ============== ========== ====================================

        Example: Two probes named *sensor* and *enable* which are 8 bits and 1 bit wide.
        ::

            pdefs = [
                {'name': 'sensor', 'net_name': 'sensor', 'map': 'port3[7:0]'},
                {'name': 'enable', 'net_name': 'enable', 'map': 'port1[3]'},
            ]
            self.define_probe(pdefs)


        Example of 'map' values:
        ::

            A 'map' value is a white space separated list of port slices and constants in any order,
             e.g. "111 probe4[7..1] 0" Constants are binary and ports are referred by port index.
            Underline character ('_') is allowed for readability in constants.
            Data only probes, can have constant bits and map multiple ports.
            Trigger ports cannot have constants bits and must map only to one port.

            "0101" - specifies a binary constant value.
            "port0[31:0]" - specifies a 32-bit slice of a port.
            "port0" - No explicit range means all bits of the physical port. "port0" means same as "probe0[31:0]", if the port is 32 bits wide.
            "port0[4]" - Shorter notation for "port0[4:4]
            "0101 port0[31:0] 1101" - specifies a constant and port slice.

        :param probe_defs: Probe definitions.
        :param done: Callback with result and any error.
        """
        service, done_cb = self.make_done(done)
        token = service.define_probe(self.ctx, probe_defs, done_cb)
        return self.add_pending(token)

    def define_port_probes(self, options: Dict[str, Any], done: DoneHWCommand) -> None:
        """
        Define ILA probe(s), one probe per ILA core port. Name of probe will be 'name_prefix' + "port_index".

        ============== ========== ============================================
        Options        Value Type Description
        ============== ========== ============================================
        name_prefix    |str|      Default: 'probe\_'.
        force          |bool|     Default: False. 'True' value will remove
                                    any previous probes with the same names.
        ============== ========== ============================================

        :param options: Property definitions.
        :param done: Callback with result and any error.
        """
        service, done_cb = self.make_done(done)
        token = service.define_port_probes(self.ctx, options, done_cb)
        return self.add_pending(token)

    def undefine_probe(self, probe_names: List[str], done: DoneHWCommand) -> None:
        """
        Delete probes.

        :param probe_names: List of probe names. Empty list means delete every probe.
        :param done: Callback with result and any error.
        """
        service, done_cb = self.make_done(done)
        token = service.undefine_probe(self.ctx, probe_names, done_cb)
        return self.add_pending(token)

    def get_probe(
        self, probe_names: List[str], attrs: List[str], done: DoneHWCommand
    ) -> Dict[str, Dict[str, Any]]:
        """
        Return probe definitions. Bit ranges of probes.

        :param probe_names: Probe name or list of probe names. Empty list means every probe.
        :param attrs: Attributes to return for probes. Empty list means all attributes.
        :return: Top dict key is probe name. Inner dict values are probe attrs :ref:`ILAProbeAttrs` .
        """
        service, done_cb = self.make_done(done)
        token = service.get_probe(self.ctx, probe_names, attrs, done_cb)
        return self.add_pending(token)

    def set_probe(self, attrs: Dict[str, Dict[str, Any]], done: DoneHWCommand) -> None:
        """

        :param attrs: Top dict key is probe name. Inner dict values are dicts with probe attributes :ref:`ILAProbeAttrs`.
        :param done: Callback with result and any error.
        """
        service, done_cb = self.make_done(done)
        token = service.set_probe(self.ctx, attrs, done_cb)
        return self.add_pending(token)

    def reset_probe(
        self, reset_trigger_values: bool, reset_capture_values: bool, done: DoneHWCommand
    ) -> None:
        """
        :param reset_trigger_values: Reset all probe trigger values to don't cares.
        :param reset_capture_values: Reset all probe capture values to don't cares.
        :param done: Callback with result and any error.
        """
        service, done_cb = self.make_done(done)
        token = service.reset_probe(self.ctx, reset_trigger_values, reset_capture_values, done_cb)
        return self.add_pending(token)

    #
    # Core Communication Methods
    #
    def arm(self, done: DoneHWCommand) -> None:
        """
        Arm the core with current *control* property settings.

        :param done: Callback with result and any error.
        """
        service, done_cb = self.make_done(done)
        token = service.arm(self.ctx, done_cb)
        return self.add_pending(token)

    def get_trigger_registers(self, done: DoneHWCommand) -> Dict[str, str]:
        """
        Get current trigger register setup as hex string numbers.
        Helpful when preparing a design for trigger-at-startup.

        :param done: Callback with result and any error.
        :return: Register hex values, for trigger setup.

        """
        service, done_cb = self.make_done(done)
        token = service.get_trigger_registers(self.ctx, done_cb)
        return self.add_pending(token)

    def upload(self, done: DoneHWCommand) -> bool:
        """
        Checks if ILA has triggered. If so, will upload waveform and set property group 'Data'
        with waveform data.

        :param done: Callback with result and any error.
        :return: True, if command uploaded a waveform.
        """
        service, done_cb = self.make_done(done)
        token = service.upload(self.ctx, done_cb)
        return self.add_pending(token)

    def monitor_status(self, status_processor_fn, max_wait_minutes=None):
        """
        Keeps checking ILA status until buffer is full or timeout.
        request.result has value None until buffer is full when it is assigned status(ILAStatus).
        request.progress is an chipscopy.ila.ILAStatus instance, with current ILA status.

        :param status_processor_fn: Function which processes status tcf properties.
        :param max_wait_minutes(float: Max number of minutes, until timeout.
        """
        assert self.request
        service = self.get_service_proxy()
        start_time = time.time()
        prev_status = None
        timeout = max_wait_minutes * 60.0 if max_wait_minutes is not None else None

        def check_status(token, error, result):
            nonlocal prev_status
            status = None
            # check if canceled
            if not self.request:
                return
            if not error:
                status = status_processor_fn(result)
                if prev_status != status:
                    prev_status = status
                    self.request.set_progress(status)

            # check for error or positive result
            if error:
                self.request.set_exception(error)
            elif (
                result
                and isinstance(result, dict)
                and (
                    result.get("is_full", False)
                    or result.get("capture_state", ILAState.IDLE) == ILAState.IDLE
                )
            ):
                self.request.set_result(status)
            elif timeout is not None and time.time() - start_time > timeout:
                if not self.request._error:
                    self.request.set_result(None)
            else:  # Buffer not yet full, check again in 500ms
                protocol.invokeLaterWithDelay(
                    500, service.refresh_property_group, self.ctx, ["status"], done=check_status
                )

        service.refresh_property_group(self.ctx, ["status"], check_status)



# *****************************************************************************
# * Copyright (c) 2011, 2013-2014, 2016 Wind River Systems, Inc. and others.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Wind River Systems - initial API and implementation
# *****************************************************************************

"""TCF Runcontrol service interface.

.. |detach| replace:: :meth:`~RunControlContext.detach`
.. |doneGetState| replace:: :meth:`~DoneGetState.doneGetState`
.. |getAddress| replace:: :meth:`~RunControlISA.getAddress`
.. |getChildren| replace:: :meth:`~RunControlService.getChildren`
.. |getContext| replace:: :meth:`~RunControlService.getContext`
.. |getISA| replace:: :meth:`~RunControlContext.getISA`
.. |getSize| replace:: :meth:`~RunControlISA.getSize`
.. |getState| replace:: :meth:`~RunControlContext.getState`
.. |resume| replace:: :meth:`~RunControlContext.resume`
.. |suspend| replace:: :meth:`~RunControlContext.suspend`
.. |terminate| replace:: :meth:`~RunControlContext.terminate`
.. |DoneGetChildren| replace:: :class:`DoneGetChildren`
.. |DoneCommand| replace:: :class:`DoneCommand`
.. |DoneGetContext| replace:: :class:`DoneGetContext`
.. |DoneGetISA| replace:: :class:`DoneGetISA`
.. |MemoryContext| replace:: :class:`~tcf.services.memory.MemoryContext`
.. |RunControlContext| replace:: :class:`RunControlContext`
.. |RunControlISA| replace:: :class:`RunControlISA`
.. |RunControlListener| replace:: :class:`RunControlListener`

Properties
----------
RunControl Context Properties
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
+--------------------+--------------+-----------------------------------------+
| Name               | Type         | Description                             |
+====================+==============+=========================================+
| PROP_BP_GROUP      | |basestring| | Context ID of a breakpoints group that  |
|                    |              | contains the context. Members of same   |
|                    |              | group share same breakpoint instances:  |
|                    |              | a breakpoint is planted once for the    |
|                    |              | group, no need to plant the breakpoint  |
|                    |              | for each member of the group.           |
+--------------------+--------------+-----------------------------------------+
| PROP_CAN_COUNT     | |int|        | Bit-set of `Resume Modes`_ values that  |
|                    |              | can be used with count > 1.             |
+--------------------+--------------+-----------------------------------------+
| PROP_CAN_DETACH    | |bool|       | **True** if detach command is supported |
|                    |              | by the context.                         |
+--------------------+--------------+-----------------------------------------+
| PROP_CAN_RESUME    | |bool|       | Bit-set of `Resume Modes`_ values that  |
|                    |              | are supported by the context.           |
+--------------------+--------------+-----------------------------------------+
| PROP_CAN_SUSPEND   | |bool|       | **True** if suspend command is supported|
|                    |              | by the context.                         |
+--------------------+--------------+-----------------------------------------+
| PROP_CAN_TERMINATE | |bool|       | **True** if terminate command is        |
|                    |              | supported by the context.               |
+--------------------+--------------+-----------------------------------------+
| PROP_CREATOR_ID    | |basestring| | ID of a context that created this       |
|                    |              | context.                                |
+--------------------+--------------+-----------------------------------------+
| PROP_HAS_STATE     | |bool|       | **True** if context has execution       |
|                    |              | state - can be suspended/resumed.       |
+--------------------+--------------+-----------------------------------------+
| PROP_ID            | |basestring| | Run control context ID.                 |
+--------------------+--------------+-----------------------------------------+
| PROP_IS_CONTAINER  | |bool|       | **True** if the context is a container. |
|                    |              | Container can propagate run control     |
|                    |              | commands to his children.               |
+--------------------+--------------+-----------------------------------------+
| PROP_NAME          | |basestring| | Human readable context name.            |
+--------------------+--------------+-----------------------------------------+
| PROP_PARENT_ID     | |basestring| | Context parent (owner) ID, for a thread |
|                    |              | it is same as process ID.               |
+--------------------+--------------+-----------------------------------------+
| PROP_PROCESS_ID    | |basestring| | Context process (memory space) ID.      |
+--------------------+--------------+-----------------------------------------+
| PROP_RC_GROUP      | |basestring| | Context ID of a run control group that  |
|                    |              | contains the context. Members of same   |
|                    |              | group are always suspended and resumed  |
|                    |              | together: resuming/suspending a context |
|                    |              | resumes/suspends all members of the     |
|                    |              | group.                                  |
+--------------------+--------------+-----------------------------------------+
| PROP_SYMBOLS_GROUP | |basestring| | Context ID of a symbols group that      |
|                    |              | contains the context. Members of a      |
|                    |              | symbols group share same symbol reader  |
|                    |              | configuration settings, like user       |
|                    |              | defined memory map entries and source   |
|                    |              | lookup info.                            |
+--------------------+--------------+-----------------------------------------+

.. _RunControl-Resume-Modes:

Resume Modes
^^^^^^^^^^^^
All resume modes are of type |int|.

+----------------------------+------------------------------------------------+
| RM_RESUME                  | Resume context.                                |
+----------------------------+------------------------------------------------+
| RM_REVERSE_RESUME          | Start running backwards. Execution will        |
|                            | continue until suspended by command or         |
|                            | breakpoint.                                    |
+----------------------------+------------------------------------------------+
| RM_REVERSE_STEP_INTO       | Reverse of ``RM_STEP_INTO``. This effectively  |
|                            | "un-executes" the previous instruction.        |
+----------------------------+------------------------------------------------+
| RM_REVERSE_STEP_INTO_LINE  | Reverse of ``RM_STEP_INTO_LINE``, Resume       |
|                            | backward execution of given context until      |
|                            | control reaches an instruction that belongs to |
|                            | a different line of source code. If a function |
|                            | is called, stop at the beginning of the last   |
|                            | line of the function code. Error is returned if|
|                            | line number information not available.         |
+----------------------------+------------------------------------------------+
| RM_REVERSE_STEP_INTO_RANGE | Reverse of ``RM_STEP_INTO_RANGE``.             |
+----------------------------+------------------------------------------------+
| RM_REVERSE_STEP_OUT        | Reverse of ``RM_STEP_OUT``. Resume backward    |
|                            | execution of the given context until control   |
|                            | reaches the point where the current function   |
|                            | was called.                                    |
+----------------------------+------------------------------------------------+
| RM_REVERSE_STEP_OVER       | Reverse of ``RM_STEP_OVER`` - run backwards    |
|                            | over a single instruction. If the instruction  |
|                            | is a function call then don't stop until get   |
|                            | out of the function.                           |
+----------------------------+------------------------------------------------+
| RM_REVERSE_STEP_OVER_LINE  | Reverse of ``RM_STEP_OVER_LINE``. Resume       |
|                            | backward execution of given context until      |
|                            | control reaches an instruction that belongs to |
|                            | a different source line. If the line contains  |
|                            | a function call then don't stop until get out  |
|                            | of the function. Error is returned if line     |
|                            | number information not available.              |
+----------------------------+------------------------------------------------+
| RM_REVERSE_STEP_OVER_RANGE | Reverse of ``RM_STEP_OVER_RANGE``.             |
+----------------------------+------------------------------------------------+
| RM_REVERSE_UNTIL_ACTIVE    | Run reverse until the context becomes active.  |
+----------------------------+------------------------------------------------+
| RM_STEP_INTO               | Step a single instruction. If the instruction  |
|                            | is a function call then stop at first          |
|                            | instruction of the function.                   |
+----------------------------+------------------------------------------------+
| RM_STEP_INTO_LINE          | Step a single source code line.  If the line   |
|                            | contains a function call then stop at first    |
|                            | line of the function.                          |
+----------------------------+------------------------------------------------+
| RM_STEP_INTO_RANGE         | Step instruction until PC is outside the       |
|                            | specified range for any reason.                |
+----------------------------+------------------------------------------------+
| RM_STEP_OUT                | Run until control returns from current         |
|                            | function.                                      |
+----------------------------+------------------------------------------------+
| RM_STEP_OVER               | Step over a single instruction. If the         |
|                            | instruction is a function call then don't stop |
|                            | until the function returns.                    |
+----------------------------+------------------------------------------------+
| RM_STEP_OVER_LINE          | Step over a single source code line. If the    |
|                            | line contains a function call then don't stop  |
|                            | until the function returns.                    |
+----------------------------+------------------------------------------------+
| RM_STEP_OVER_RANGE         | Step over instructions until PC is outside the |
|                            | specified range. Any function call within the  |
|                            | range is considered to be in range.            |
+----------------------------+------------------------------------------------+
| RM_UNTIL_ACTIVE            | Run until the context becomes active -         |
|                            | scheduled to run on a target CPU.              |
+----------------------------+------------------------------------------------+

State Reasons
^^^^^^^^^^^^^
State change reason of a context. Reason can be any text, but if it is one of
predefined strings, a generic client might be able to handle it better.

+---------------------+-------------------------------------------------------+
| Name                | Description                                           |
+=====================+=======================================================+
| REASON_BREAKPOINT   | Context has been suspended by a breakpoint hit.       |
+---------------------+-------------------------------------------------------+
| REASON_CONTAINER    | Context is supspended because it container is         |
|                     | suspended.                                            |
+---------------------+-------------------------------------------------------+
| REASON_ERROR        | Context has been suspended by an error.               |
+---------------------+-------------------------------------------------------+
| REASON_EXCEPTION    | Context has been suspended by an exception.           |
+---------------------+-------------------------------------------------------+
| REASON_SHAREDLIB    | Context has been suspended by a sharedlib.            |
+---------------------+-------------------------------------------------------+
| REASON_SIGNAL       | Context has been suspended by a signal.               |
+---------------------+-------------------------------------------------------+
| REASON_STEP         | Context has been suspended by a step request.         |
+---------------------+-------------------------------------------------------+
| REASON_USER_REQUEST | Context has been suspended by user request.           |
+---------------------+-------------------------------------------------------+
| REASON_WATCHPOINT   | Context has been suspended by a watchpoint hit.       |
+---------------------+-------------------------------------------------------+

State Optional Parameters
^^^^^^^^^^^^^^^^^^^^^^^^^
Those information may appear in the context state. See the *param* parameter
of |doneGetState|.

+--------------------------+--------------+-----------------------------------+
| Name                     | Type         | Description                       |
+==========================+==============+===================================+
| STATE_BREAKPOINT_IDS     | |list|       | Breakpoint IDs the context is     |
|                          |              | suspended on.                     |
+--------------------------+--------------+-----------------------------------+
| STATE_PC_ERROR           | |int|        | Program counter the error occurred|
|                          |              | at.                               |
+--------------------------+--------------+-----------------------------------+
| STATE_REVERSING          | |bool|       | true if the context is running in |
|                          |              | reverse.                          |
+--------------------------+--------------+-----------------------------------+
| STATE_SIGNAL             | | int|       | Signal number.                    |
+--------------------------+--------------+-----------------------------------+
| STATE_SIGNAL_DESCRIPTION | |basestring| | Signal description.               |
+--------------------------+--------------+-----------------------------------+
| STATE_SIGNAL_NAME        | |basestring| | Signal name.                      |
+--------------------------+--------------+-----------------------------------+

Resume Optional Parameters
^^^^^^^^^^^^^^^^^^^^^^^^^^
Some resume modes may need optional parameters. See *params* parameter of the
|resume| command.

All resume optional paramters are of type |int|.

+----------------+------------------------------------------------------------+
| Name           | Description                                                |
+================+============================================================+
| RP_RANGE_END   | Ending address of step range, exclusive.                   |
+----------------+------------------------------------------------------------+
| RP_RANGE_START | Starting address of step range, inclusive.                 |
+----------------+------------------------------------------------------------+

Instruction Set Architecture Properties
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

+--------------------------+--------------+-----------------------------------+
| Name                     | Type         | Description                       |
+==========================+==============+===================================+
| ISA_ADDRESS              | |int|        | The address the Instruction Set   |
|                          |              | Architecture is valid at.         |
+--------------------------+--------------+-----------------------------------+
| ISA_ALIGNMENT            | |int|        | Instruction alignmement for the   |
|                          |              | Instruction Set Architecture.     |
+--------------------------+--------------+-----------------------------------+
| ISA_DEFAULT              | |basestring| | Name of the default ISA.          |
+--------------------------+--------------+-----------------------------------+
| ISA_MAX_INSTRUCTION_SIZE | |int|        | Maximum instruction size for ISA. |
+--------------------------+--------------+-----------------------------------+
| ISA_NAME                 | |basestring| | Name of the ISA.                  |
+--------------------------+--------------+-----------------------------------+
| ISA_SIZE                 | |int|        | Size of the ISA. Another ISA may  |
|                          |              | be defined after                  |
|                          |              | ``ISA_ADDRESS`` + ``ISA_SIZE``    |
+--------------------------+--------------+-----------------------------------+


Service Methods
---------------
.. autodata:: NAME
.. autoclass:: RunControlService

addListener
^^^^^^^^^^^
.. automethod:: RunControlService.addListener

getChildren
^^^^^^^^^^^
.. automethod:: RunControlService.getChildren

getContext
^^^^^^^^^^
.. automethod:: RunControlService.getContext

getName
^^^^^^^
.. automethod:: RunControlService.getName

removeListener
^^^^^^^^^^^^^^
.. automethod:: RunControlService.removeListener

detach
^^^^^^
.. automethod:: RunControlContext.detach

getISA
^^^^^^
.. automethod:: RunControlContext.getISA

getState
^^^^^^^^
.. automethod:: RunControlContext.getState

resume
^^^^^^
.. automethod:: RunControlContext.resume

suspend
^^^^^^^
.. automethod:: RunControlContext.suspend

terminate
^^^^^^^^^
.. automethod:: RunControlContext.terminate

Callback Classes
----------------
DoneCommand
^^^^^^^^^^^
.. autoclass:: DoneCommand
    :members:

DoneGetChildren
^^^^^^^^^^^^^^^
.. autoclass:: DoneGetChildren
    :members:

DoneGetContext
^^^^^^^^^^^^^^
.. autoclass:: DoneGetContext
    :members:

DoneGetISA
^^^^^^^^^^
.. autoclass:: DoneGetISA
    :members:

DoneGetState
^^^^^^^^^^^^
.. autoclass:: DoneGetState
    :members:

Listeners
---------
RunControlListener
^^^^^^^^^^^^^^^^^^
.. autoclass:: RunControlListener
    :members:

Helper Classes
--------------
RunControlContext
^^^^^^^^^^^^^^^^^
.. autoclass:: RunControlContext
    :members:

RunControlError
^^^^^^^^^^^^^^^
.. autoclass:: RunControlError
    :members:
    :show-inheritance:

RunControlISA
^^^^^^^^^^^^^
.. autoclass:: RunControlISA
    :members:
"""

from .. import services

NAME = "RunControl"
"""RunControl service name."""

# Context property names.

PROP_ID = "ID"
PROP_PARENT_ID = "ParentID"
PROP_PROCESS_ID = "ProcessID"
PROP_CREATOR_ID = "CreatorID"
PROP_NAME = "Name"
PROP_IS_CONTAINER = "IsContainer"
PROP_HAS_STATE = "HasState"
PROP_CAN_RESUME = "CanResume"
PROP_CAN_COUNT = "CanCount"
PROP_CAN_SUSPEND = "CanSuspend"
PROP_CAN_TERMINATE = "CanTerminate"
PROP_RC_GROUP = "RCGroup"
PROP_BP_GROUP = "BPGroup"
PROP_CAN_DETACH = "CanDetach"
PROP_SYMBOLS_GROUP = "SymbolsGroup"

# Context resume modes.

RM_RESUME = 0
RM_STEP_OVER = 1
RM_STEP_INTO = 2
RM_STEP_OVER_LINE = 3
RM_STEP_INTO_LINE = 4
RM_STEP_OUT = 5
RM_REVERSE_RESUME = 6
RM_REVERSE_STEP_OVER = 7
RM_REVERSE_STEP_INTO = 8
RM_REVERSE_STEP_OVER_LINE = 9
RM_REVERSE_STEP_INTO_LINE = 10
RM_REVERSE_STEP_OUT = 11
RM_STEP_OVER_RANGE = 12
RM_STEP_INTO_RANGE = 13
RM_REVERSE_STEP_OVER_RANGE = 14
RM_REVERSE_STEP_INTO_RANGE = 15
RM_UNTIL_ACTIVE = 16
RM_REVERSE_UNTIL_ACTIVE = 17

# Suspended reasons

REASON_USER_REQUEST = "Suspended"
REASON_STEP = "Step"
REASON_BREAKPOINT = "Breakpoint"
REASON_EXCEPTION = "Exception"
REASON_CONTAINER = "Container"
REASON_WATCHPOINT = "Watchpoint"
REASON_SIGNAL = "Signal"
REASON_SHAREDLIB = "Shared Library"
REASON_ERROR = "Error"
REASON_ACTIVE = "Active"

# Optional parameters of context state.

STATE_SIGNAL = "Signal"
STATE_SIGNAL_NAME = "SignalName"
STATE_SIGNAL_DESCRIPTION = "SignalDescription"
STATE_BREAKPOINT_IDS = "BPs"
STATE_PC_ERROR = "PCError"
STATE_REVERSING = "Reversing"

# Optional parameters of resume command.

RP_RANGE_START = "RangeStart"
RP_RANGE_END = "RangeEnd"

# Instruction Set Architecture properties

ISA_ADDRESS = "Addr"
ISA_ALIGNMENT = "Alignment"
ISA_DEFAULT = "DefISA"
ISA_MAX_INSTRUCTION_SIZE = "MaxInstrSize"
ISA_NAME = "ISA"
ISA_SIZE = "Size"


class RunControlService(services.Service):
    """RunControl service interface."""

    def getName(self):
        """Get this service name.

        :returns: The value of string :const:`NAME`
        """
        return NAME

    def getContext(self, contextID, done):
        """Retrieve context properties for given context ID.

        :param contextID: ID of the context to retrieve.
        :type contextID: |basestring|
        :param done: Callback interface called when operation is completed.
        :type done: |DoneGetContext|

        :returns: Pending command handle.
        """
        raise NotImplementedError("Abstract method")

    def getChildren(self, parent_context_id, done):
        """Retrieve children of given context.

        :param parent_context_id: parent context ID. Can be **None** - to
                                  retrieve top level of the hierarchy, or one
                                  of context IDs retrieved by previous
                                  |getChildren| commands.
        :type parent_context_id: |basestring| or **None**
        :param done: Callback interface called when operation is completed.
        :type done: |DoneGetChildren|

        :returns: Pending command handle.
        """
        raise NotImplementedError("Abstract method")

    def addListener(self, listener):
        """Add run control event listener.

        :param listener: RunControl event listener to add.
        :type listener: |RunControlListener|

        :returns: **None**, always.
        """
        raise NotImplementedError("Abstract method")

    def removeListener(self, listener):
        """Remove run control event listener.

        :param listener: run control event listener to remove.
        :type listener: |RunControlListener|
        """
        raise NotImplementedError("Abstract method")


class RunControlError(Exception):
    """RunControl error class interface."""
    pass


class RunControlISA(object):
    """Instruction Set Architecture.

    :param props: The properties to initialise object with. See
                  `Instruction Set Architecture Properties`_.
    :type props: |dict|
    """
    def __init__(self, props):
        self._properties = props or {}

    def __repr__(self):
        return self.__class__.__name__ + '(' + str(self.getProperties()) + ')'

    def __str__(self):
        res = self.__class__.__name__ + ' ['
        res += 'Name=' + str(self.getName()) + ', '
        res += 'Address=' + str(self.getAddress()) + ', '
        res += ISA_SIZE + '=' + str(self.getSize()) + ', '
        res += ISA_ALIGNMENT + '=' + str(self.getAlignment()) + ', '
        res += ISA_MAX_INSTRUCTION_SIZE + '='
        res += str(self.getMaxInstructionSize()) + ', '
        res += 'Default=' + str(self.getDefault())
        res += ']'
        return res

    def getAddress(self):
        """Get the address this ISA is defined at.

        :returns: An |int| representing the address at which this ISA is
                  defined, or **0** if it is not known.
        """
        return self._properties.get(ISA_ADDRESS, 0)

    def getAlignment(self):
        """Get the address alignment for this ISA.

        :returns: An |int| representing the address alignment for this ISA, or
                  **0** if it is not known.
        """
        return self._properties.get(ISA_ALIGNMENT, 0)

    def getDefault(self):
        """Get the default name of this ISA.

        :returns: A |basestring| representing the default name of this ISA,
                  or **None** if it is not known.
        """
        return self._properties.get(ISA_DEFAULT, None)

    def getMaxInstructionSize(self):
        """Get the maximum size of an instruction for this ISA.

        :returns: An |int| representing the maximum instruction size for this
                  ISA, or **0** if it is not known.
        """
        return self._properties.get(ISA_MAX_INSTRUCTION_SIZE, 0)

    def getName(self):
        """Get the name of this ISA.

        :returns: A |basestring| representing the name of this ISA, or **None**
                  if it is not known.
        """
        return self._properties.get(ISA_NAME, None)

    def getProperties(self):
        """Get the properties defining this ISA.

        :returns: A |dict| representing all the defined properties for this
                  ISA.
        """
        return self._properties

    def getSize(self):
        """Get the size of this ISA.

        Starting from |getAddress|, this ISA is valid only for |getSize| bytes.
        Another |getISA| request should be performed for further addresses.

        :returns: An |int| representing the size for this ISA, or **0** if it
                  is not known.
        """
        return self._properties.get(ISA_MAX_INSTRUCTION_SIZE, 0)


class DoneGetISA(object):
    """Client call back interface for |getISA|."""
    def doneGetISA(self, token, error, isa):
        """Called when |getISA| command execution is complete.

        :param token: Pending command handle.
        :param error: Command execution error or **None**.
        :param isa: A |RunControlISA| object result of the request.
        :type isa: |RunControlISA|
        """
        pass


class DoneGetState(object):
    """Client call back interface for |getState|."""
    def doneGetState(self, token, error, suspended, pc, reason, params):
        """Called when |getState| command execution is complete.

        :param token: Pending command handle.
        :param error: Pommand execution error or **None**.
        :param suspended: **True** if the context is suspended.
        :type suspended: |bool|
        :param pc: Program counter of the context (if suspended).
        :type pc: |int|
        :param reason: Suspend reason (if suspended), see `State Reasons`_.
        :type reason: |basestring|
        :param params: Additional target specific data about context state,
                       see `State Optional Parameters`_.
        :type params: |dict|
        """
        pass


class DoneCommand():
    """Client call back interface for |detach|, |resume|, |suspend| and
    |terminate|."""

    def doneCommand(self, token, error):
        """Called when run control command execution is complete.

        :param token: Pending command handle.
        :param error: Command execution error or **None**.
        """
        pass


class DoneGetContext():
    "Client callback interface for |getContext|."

    def doneGetContext(self, token, error, context):
        """Called when context data retrieval is done.

        :param token: Pending command handle.
        :param error: Error description if operation failed, **None** if
                      succeeded.
        :param context: Context data.
        :type context: |RunControlContext|
        """
        pass


class DoneGetChildren(object):
    """Client callback interface for |getChildren|."""

    def doneGetChildren(self, token, error, context_ids):
        """Called when context list retrieval is done.

        :param token: Pending command handle.
        :param error: Error description if operation failed, **None** if
                      succeeded.
        :param context_ids: array of available context IDs.
        :type context_ids: |list|
        """
        pass


class RunControlContext(object):
    """A context corresponds to an execution thread, process, address space,
    etc.

    A context can belong to a parent context. Contexts hierarchy can be simple
    plain list or it can form a tree. It is up to target agent developers to
    choose layout that is most descriptive for a given target. Context IDs are
    valid across all services. In other words, all services access same
    hierarchy of contexts, with same IDs, however, each service accesses its
    own subset of context's attributes and functionality, which is relevant to
    that service.

    :param props: a `dict` of properties to initialise this RunControl context
                  with. See all the `RunControl Context Properties`_ constants.
    """
    def __init__(self, props):
        self._props = props or {}

    def __str__(self):
        return "[Run Control Context %s]" % self._props

    def getProperties(self):
        """Get context properties.

        See `RunControl Context Properties`_ definitions for property names.

        Context properties are read only, clients should not try to modify
        them.

        :returns: A |dict| of context properties.
        """
        return self._props

    def getID(self):
        """Retrieve context ID.

        :returns: A |basestring| representing this RunControl context ID.
        """
        return self._props.get(PROP_ID)

    def getParentID(self):
        """Retrieve parent context ID.

        :returns: A |basestring| representing this RunControl parent context
                  ID or **None**.
        """
        return self._props.get(PROP_PARENT_ID)

    def getProcessID(self):
        """Retrieve context process ID.

        :returns: A |basestring| representing this RunControl process ID. This
                  value is also meant to be the ID of the |MemoryContext|
                  this RunControl context belongs to.
        """
        return self._props.get(PROP_PROCESS_ID)

    def getCreatorID(self):
        """Retrieve context creator ID.

        :returns: A |basestring| representing this RunControl creator context
                  ID or **None**.
        """
        return self._props.get(PROP_CREATOR_ID)

    def getName(self):
        """Retrieve human readable context name.

        :returns: A |basestring| representing this RunControl context name or
                  **None**
        """
        return self._props.get(PROP_NAME)

    def isContainer(self):
        """Check if this context is a RunControl Context container.

        Executing resume or suspend command on a container causes all its
        children to resume or suspend.

        :returns: A |bool| stating if this context is a container or not.
        """
        return self._props.get(PROP_IS_CONTAINER)

    def hasState(self):
        """Check if this context has a state.

        Only context that has a state can be resumed or suspended.

        :returns: A |bool| stating if this context has a state or not.
        """
        return self._props.get(PROP_HAS_STATE)

    def canSuspend(self):
        """Check if this context can be suspended.

        Value **True** means suspend command is supported by the context,
        however the method does not check that the command can be executed
        successfully in the current state of the context. For example, the
        command still can fail if context is already suspended.

        :return: A |bool| stating if this context can be suspended or not.
        """
        return self._props.get(PROP_CAN_SUSPEND)

    def canResume(self, mode):
        """Check if this context can resume with given *mode*.

        Value **True** means resume command is supported by the context,
        however the method does not check that the command can be executed
        successfully in the current state of the context. For example, the
        command still can fail if context is already resumed.

        :param mode: Resume mode. See `Resume Modes`_.
        :type mode: |int|

        :returns: A |bool| stating if this context can resume with given
                  *mode*.
        """
        b = self._props.get(PROP_CAN_RESUME) or 0
        return (b & (1 << mode)) != 0

    def canCount(self, mode):
        """Check if context can count resumes of given *mode*.

        Value **True** means resume command with count other than 1 is
        supported by the context, however the method does not check that the
        command can be executed successfully in the current state of the
        context. For example, the command still can fail if context is already
        resumed.

        :param mode: Resume mode. See `Resume Modes`_.
        :type mode: |int|

        :returns: A |bool| stating if this context can count resumes on given
                  *mode*.
        """
        b = self._props.get(PROP_CAN_COUNT) or 0
        return (b & (1 << mode)) != 0

    def canTerminate(self):
        """Check if context can be terminated.

        Value **True** means terminate command is supported by the context,
        however the method does not check that the command can be executed
        successfully in the current state of the context. For example, the
        command still can fail if context is already exited.

        :returns: A |bool| stating if this context can be terminated.
        """
        return self._props.get(PROP_CAN_TERMINATE)

    def getRCGroup(self):
        """Get the context RunControl Group ID.

        The RunContol Group ID is the context ID of a run control group that
        contains the context.

        Members of same group are always suspended and resumed together:
        resuming/suspending a context resumes/suspends all members of the
        group.

        :returns: A |basestring| representing this context RC group or
                  **None**.
        """
        return self._props.get(PROP_RC_GROUP)

    def getBPGroup(self):
        """Get this context Breakpoints group ID.

        The Breakpoints group ID is the context ID of a breakpoints group that
        contains the context.

        Members of same group share same breakpoint instances: a breakpoint is
        planted once for the group, no need to plant the breakpoint for each
        member of the group

        :returns: A |basestring| representing this context Breakpoints group
                  ID, or **None** if the context does not support breakpoints.
        """
        return self._props.get(PROP_BP_GROUP)

    def getSymbolsGroup(self):
        """Get this context Symbols group ID.

        The Symbols group ID is the context ID of a symbols group that contains
        the context.

        Members of a symbols group share same symbol reader configuration
        settings, like user defined memory map entries and source lookup info.

        :returns: A |basestring| representing this context Symbols group ID, or
                  **None** if the context is not a member of a symbols group.
        """
        return self._props.get(PROP_SYMBOLS_GROUP)

    def canDetach(self):
        """Check if this context can be detached.

        Value **True** means detach command is supported by the context,
        however the method does not check that the command can be executed
        successfully in the current state of the context. For example, the
        command still can fail if the context already has exited.

        :returns: A |bool| stating if this context can be detached or not.
        """
        return self._props.get(PROP_CAN_DETACH)

    def detach(self, done):
        """Send a command to detach a context.

        :param done: Command result call back object
        :type done: |DoneCommand|

        :returns: Pending command handle, can be used to cancel the command.
        """
        raise NotImplementedError("Abstract method")

    def getISA(self, address, done):
        """Send a command to retrieve current Instruction Set Architecture of a
        context.

        :param address: The memory address to get ISA for.
        :type address: |int|
        :param done: Command result call back object.
        :type done: |DoneGetISA|

        :returns: Pending command handle, can be used to cancel the command.
        """
        raise NotImplementedError("Abstract method")

    def getState(self, done):
        """Send a command to retrieve current state of a context.

        :param done: Command result call back object.
        :type done: |DoneGetState|

        :returns: Pending command handle, can be used to cancel the command.
        """
        raise NotImplementedError("Abstract method")

    def suspend(self, done):
        """Send a command to suspend a context.

        Also suspends children if context is a container.

        :param done: Command result call back object.
        :type done: |DoneCommand|

        :returns: Pending command handle, can be used to cancel the command.
        """
        raise NotImplementedError("Abstract method")

    def resume(self, mode, count, params, done):
        """Send a command to resume a context.

        Also resumes children if context is a container.

        :param mode: Defines how to resume the context. See `Resume Modes`.
        :type mode: |int|
        :param count: If mode implies stepping, defines how many steps to
                      perform.
        :type count: |int|
        :param params: Resume parameters, for example, step range definition,
                       see `Resume Optional Parameters`_.
        :type params: |dict|
        :param done: command result call back object
        :type done: |DoneCommand|

        :returns: Pending command handle, can be used to cancel the command.
        """
        raise NotImplementedError("Abstract method")

    def terminate(self, done):
        """Send a command to terminate a context.

        :param done: Command result call back object.
        :type done: |DoneCommand|

        :returns: Pending command handle, can be used to cancel the command.
        """
        raise NotImplementedError("Abstract method")


class RunControlListener(object):
    """RunControl service events listener interface."""

    def contextAdded(self, contexts):
        """Called when new contexts are created.

        :param contexts: List of new context properties. A |RunControlContext|
                         can be initialised with those |dict| properties.
        :type contexts: |list| of |dict|
        """
        pass

    def contextChanged(self, contexts):
        """Called when a context properties changed.

        :param contexts: List of new context properties. A `RunControlContext`
                         can be initialised with those |dict| properties.
        :type contexts: |list| of |dict|
        """
        pass

    def contextRemoved(self, context_ids):
        """Called when contexts are removed.

        :param context_ids: List of removed context IDs.
        :type context_ids: |list|
        """
        pass

    def contextStateChanged(self, context):
        """Called when context state changes and the context is not and was not
        in suspended state.

        Changes to and from suspended state should be reported by other events:
        contextSuspended, contextResumed, containerSuspended, containerResumed.

        :param context: ID of a context that changed state.
        :type context: |basestring|
        """
        pass

    def contextSuspended(self, context, pc, reason, params):
        """Called when a thread is suspended.

        :param context: ID of a context that was suspended.
        :type context: |basestring|
        :param pc: Program counter of the context.
        :type pc: |int| or **None**
        :param reason: Human readable description of suspend reason. See
                       `State Reasons`_.
        :type reason: |basestring|
        :param params: Additional, target specific data about suspended
                       context. See `State Optional Parameters`_.
        :type params: |dict|
        """
        pass

    def contextResumed(self, context):
        """Called when a thread is resumed.

        :param context: ID of a context that was resumed.
        :type context: |basestring|
        """
        pass

    def containerSuspended(self, context, pc, reason, params, suspended_ids):
        """Called when target simultaneously suspends multiple threads in a
        container (process, core, etc.).

        :param context: ID of a context responsible for the event. It can be
                        container ID or any one of container children, for
                        example, it can be thread that hit "suspend all"
                        breakpoint. Client expected to move focus (selection)
                        to this context.
        :type context: |basestring|
        :param pc: program counter of the context.
        :type pc: |int|
        :param reason: Suspend reason, see `State Reasons`_.
        :type reason: |basestring|
        :param params: Additional target specific data about context state,
                       see `State Reasons`_.
        :type params: |dict|
        :param suspended_ids: full list of all contexts that were suspended.
        :type suspended_ids: |list|
        """
        pass

    def containerResumed(self, context_ids):
        """Called when target simultaneously resumes multiple threads in a
        container (process, core, etc.).

        :param context_ids: Full list of all contexts that were resumed.
        :type context_ids: |list|
        """
        pass

    def contextException(self, context, msg):
        """Called when an exception is detected in a target thread.

        :param context: ID of a context that caused an exception.
        :type context: |basestring|
        :param msg: Human readable description of the exception.
        :type msg: |basestring|
        """
        pass



# *****************************************************************************
# * Copyright (c) 2015-2016 Wind River Systems, Inc. and others.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Wind River Systems - initial API and implementation
# *****************************************************************************

import sys
PY3 = sys.version_info[0] == 3


if PY3:
    ints = (int,)
    strings = (str,)
    inttype = int
    longtype = int

    def str2bytes(data):
        if data and isinstance(data[0], str):
            return bytes([ord(c) for c in data])
        elif isinstance(data, bytes):
            return data
        return bytes(data)
else:
    ints = (int, long)  # @UndefinedVariable
    strings = (basestring, str)  # @UndefinedVariable
    inttype = int
    longtype = long  # @UndefinedVariable

    def str2bytes(data):
        if data and isinstance(data[0], str):
            return bytearray([ord(c) for c in data])
        elif isinstance(data, bytearray):
            return data
        return bytearray(data)


def bytes2str(data):
    if data and isinstance(data[0], int):
        return ''.join(chr(b) for b in data)
    return str(data)



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING

from chipscopy.api.ibert.aliases import (
    ALIAS_DICT,
    HANDLE_NAME,
    MODIFIABLE_ALIASES,
    PLL_LOCK_STATUS,
    PROPERTY_ENDPOINT,
    TYPE,
    PLL_RESET,
)
from chipscopy.api.ibert.serial_object_base import SerialObjectBase
from typing_extensions import final

if TYPE_CHECKING:  # pragma: no cover
    from chipscopy.api.ibert.gt_group import GT


@final
class PLL(SerialObjectBase["GTGroup", None]):
    def __init__(self, name, parent, tcf_node, configuration):
        SerialObjectBase.__init__(
            self,
            name=name,
            type=configuration[TYPE],
            parent=parent,
            handle=configuration[HANDLE_NAME],
            core_tcf_node=tcf_node,
            property_for_alias=configuration.get(ALIAS_DICT, dict()),
            is_property_endpoint=configuration.get(PROPERTY_ENDPOINT, False),
            modifiable_aliases=configuration.get(MODIFIABLE_ALIASES, set()),
        )

        # This is used by the filter_by method in QueryList
        self.filter_by = {"name": self.name, "type": self.type, "handle": self.handle}

    def __repr__(self) -> str:
        return self.name

    @staticmethod
    def check_for_filter_match(filters_dict, filter_name, match_value) -> bool:
        if filter_name not in filters_dict:
            return False

        # Special case for "type" filter, since it can be "PLL/<RPLL/LCPLL>".
        # So use startswith instead of ==
        return (
            filters_dict[filter_name].startswith(match_value)
            if filter_name == "type"
            else filters_dict[filter_name] == match_value
        )

    @property
    def locked(self) -> bool:
        _, prop_value = self.property.refresh(self.property_for_alias[PLL_LOCK_STATUS]).popitem()
        return True if prop_value == "Locked" else False

    def reset(self):
        """
        Reset the PLL
        """
        if PLL_RESET not in self.property_for_alias:
            raise RuntimeError(f"PLL '{self.handle}' cannot be reset!")

        props = {self.property_for_alias[PLL_RESET]: 0x1}
        self.property.set(**props)
        self.property.commit(self.property_for_alias[PLL_RESET])



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading
from socket import AF_INET, SOCK_STREAM, socket, SHUT_RD
from typing import ByteString, Iterable

from chipscopy.utils.logger import log
from chipscopy.utils import bytes_from_words


def recv(sock, size=1):
    c = sock.recv(size)
    if len(c) == 0:  # pragma: no cover
        raise Exception("Disconnected")
    return c


def get_uint_le(sock, size):
    buf = recv(sock, size)
    value = 0
    for byte in buf[size - 1 :: -1]:
        value = (value << 8) | byte
    return value


def set_uint_le(value, size):
    buf = bytearray(size)
    for i in range(size):
        buf[i] = value & 0xFF
        value >>= 8
    return buf


def get_uleb128(sock):
    value = 0
    b = 0
    while True:
        n = recv(sock)[0]
        value |= (n & 0x7F) << b
        b += 7
        if (n & 0x80) == 0:
            break
    return value


def write_uleb128(value):
    buf = bytearray()
    if not value:
        buf = b"\x00"
    while value:
        if value >= 0x80:
            p = (value & 0x7F) | 0x80
        else:
            p = value & 0x7F
        value >>= 7
        buf.append(p)
    return buf


def get_cmd(sock):
    cmd_name = bytearray()
    while True:
        c = recv(sock)
        if c == b":":
            break
        cmd_name += c
    return cmd_name


class XVCPendingError(Exception):
    pass


class XVCHandler(object):
    default_capabilities = [b"memory", b"stop", b"idcode=0x4CA8093", b"idcode2=1"]

    def __init__(
        self,
        addr: int = 0,
        length: int = 0,
        sub_handlers: Iterable["XVCHandler"] = (),
        idcode: int = 0,
        idcode2: int = 0,
    ):
        self.config = {}
        self.addr = addr
        self.length = length
        self.sub_handlers = [h for h in sorted(sub_handlers, key=lambda x: x.addr) if h != self]
        self.idcode = idcode
        self.idcode2 = idcode2

    @property
    def capabilities(self):
        cap = list()
        if self.idcode:
            cap.append(str(f"idcode={self.idcode}").encode("utf-8"))
        if self.idcode2:
            cap.append(str(f"idcode2={self.idcode2}").encode("utf-8"))
        return self.default_capabilities + cap

    def add_sub_handler(self, sub_handler: "XVCHandler"):
        if sub_handler == self:
            return
        for i, h in enumerate(self.sub_handlers):
            if sub_handler.addr < h.addr:
                self.sub_handlers.insert(i, sub_handler)
                return

        self.sub_handlers.append(sub_handler)

    def set_config(self, config_str: str):
        configs = config_str.split(",")
        for config in configs:
            if config[-1] == "+":
                self.config[config[:-1]] = True
            elif config[-1] == "-":
                self.config[config[:-1]] = False
            else:
                name, sep, value = config.partition("=")
                if sep:
                    self.config[name] = value

    def is_valid_addr(self, addr: int):
        return self.addr <= addr < self.addr + self.length

    def read_mem(self, flags: int, addr: int, num_bytes: int) -> ByteString:
        data = bytearray()
        i = iter(self.sub_handlers)
        try:
            handler = next(i)

            while len(data) < num_bytes:
                cur_addr = addr + len(data)
                while handler.addr + handler.length <= addr:
                    handler = next(i)
                if handler.is_valid_addr(cur_addr):
                    loc_addr = cur_addr - handler.addr
                    length = min(num_bytes - len(data), handler.length - loc_addr)
                    data += handler.read_mem(flags, loc_addr, length)
                else:
                    length = min(handler.addr - addr, num_bytes - len(data))
                    data += self.get_mem(flags, cur_addr, length)
        except StopIteration:
            pass
        if len(data) < num_bytes:
            data += self.get_mem(flags, addr + len(data), num_bytes - len(data))

        return data

    def write_mem(self, flags: int, addr: int, data: ByteString):
        offset = 0
        num_bytes = len(data)
        i = iter(self.sub_handlers)
        try:
            handler = next(i)

            while offset < num_bytes:
                while handler.addr + handler.length <= addr:
                    handler = next(i)

                if handler.is_valid_addr(addr):
                    length = min(num_bytes - offset, handler.length - (addr - handler.addr))
                    handler.write_mem(flags, addr - handler.addr, data[offset:length])
                else:
                    length = min(handler.addr - addr, num_bytes - offset)
                    self.set_mem(flags, addr + offset, data[offset:length])
                offset += length
        except StopIteration:
            pass

        if offset < num_bytes:
            length = num_bytes - offset
            self.set_mem(flags, addr + offset, data[offset:length])

    def get_mem(self, flags: int, addr: int, num_bytes: int) -> ByteString:
        """
        Accesses memory from a given location.

        :param flags: Flag bits of memory access (currently unused)
        :param addr: Starting address of memory read
        :param num_bytes: Number of bytes to read
        :return: Data from memory access
        """
        return bytes(num_bytes)

    def set_mem(self, flags: int, addr: int, data: ByteString):
        """
        Writes to given memory location
        :param flags: Flag bits of memory access (currently unused)
        :param addr: Starting address of memory write
        :param data: Data to write
        """
        raise NotImplementedError()

    def shift(self, num_bits: int, tms: ByteString, tdi: ByteString) -> ByteString:
        """
        Performs a JTAG shift transaction
        :param num_bits: Number of bits or TCKs of JTAG shift
        :param tms: JTAG TMS signal vector
        :param tdi: JTAG TDI signal vector
        :return: JTAG TDO signal vector
        """
        return tdi

    def set_tck(self, period: int) -> int:
        """
        Sets JTAG TCK period
        :param period: Desired JTAG TCK period in ns
        :return: Result JTAG TCK period in ns
        """
        return period


class XVCServer(object):
    max_packet_len = 4096

    def __init__(self, port: int = 2542, handler: XVCHandler = None):
        if not handler:
            handler = XVCHandler()
        self.handler = handler
        self.port = port
        self.thread = None
        self.sock = None
        self.listen_sock = None
        self.listening = threading.Condition()
        self.running = False
        self.pending_error = ""

    def process_getinfo(self):
        """
        Gets the XVC Service version

        XVC Format:

        Return Format: 'xvcServer_v<xvc_server_ver>:<xvc_vector_len>\n'
        """
        self.sock.send("xvcServer_v1.1:{}\n".format(self.max_packet_len).encode("utf-8"))

    def process_capabilities(self):
        """
        Gets capabilities of XVC Server

        XVC Format:

        Return Format: [length][capability list]
        [length] ULEB128 indicating length of capability list in bytes
        [capability list] comma separated list of strings
        """
        reply = b",".join(self.handler.capabilities)
        log.xvc.debug("\treply [{}]{}".format(len(reply), reply.decode("utf-8")))
        self.sock.send(write_uleb128(len(reply)) + reply)

    def process_configure(self):
        """
        Sets configuration of XVC server

        XVC Format: [length][configuration strings]

        Params:
        [length] ULEB128 indicating length of [configuration strings] in bytes
        [configuration strings] comma separated list of strings

        Configuration string format:
        name[+|-|=value] where indicates enable, - indicates disable, and =value sets configuration to value

        Return Format: [status]
        [status] error code of status (0x00 for success)
        :return:
        """
        length = get_uleb128(self.sock)
        configs = recv(self.sock, length)
        self.handler.set_config(configs.decode("utf-8"))
        self.sock.send(b"\x00")

    def process_error(self):
        """
        Returns pending error and clears error flag

        XVC Format:

        Return Format: [length][message]
        [length] ULEB128 length of [message] in bytes
        [message] UTF-8 encoded error message

        Special machine decodable errors may be returned from commands as single word all uppercase
        """
        log.xvc.debug(f"\tmessage: {self.pending_error}")
        message = self.pending_error.encode("utf-8")
        self.pending_error = ""
        self.sock.send(write_uleb128(len(message)) + message)

    def process_shift(self):
        """
        Shifts in num bits using the byte vectors tms and tdi

        XVC Format: [num bits][tms vector][tdi vector]

        Params:
        [num_bits] little-ending (LE) number of TCK toggles and bit length of vectors
        [tms] byte sized vector with all the TMS shift bits.  Bit 0 in byte 0 is shifted out first.
        [tdi] like tms, represents all the TDI vectors to be shifted in

        Return Format: [tdo vector]
        [tdo] TDO vector bits for every bit shifted in.  Bit 0 in byte 0 is shifted out first.
        """
        num_bits = get_uint_le(self.sock, 4)
        num_bytes = int((num_bits + 7) / 8)
        tms = recv(self.sock, num_bytes)
        tdi = recv(self.sock, num_bytes)
        self.sock.send(self.handler.shift(num_bits, tms, tdi))

    def process_settck(self):
        """
        Attempts to set the JTAG TCK period

        XVC Format: [period]

        Params:
        [period] LE TCK period in ns

        Return Format: [period]
        [period] LE value of resulting TCK period in ns
        """
        period = get_uint_le(self.sock, 4)
        log.xvc.debug(f"\tperiod: {period}")
        self.sock.send(set_uint_le(self.handler.set_tck(period), 4))

    def process_mwr(self):
        """
        XVC Format: [flags][addr][num bytes][buffer]

        Params:
        [flags] ULEB128 indicating flag bits (currently unused)
        [addr] ULEB128 starting address for mem write
        [num bytes] ULEB128 number of bytes in buffer
        [buffer] byte buffer of data to write

        Return Format: [status]
        [status] error code of status (0x00 for success)
        """
        flags = get_uleb128(self.sock)
        addr = get_uleb128(self.sock)
        num_bytes = get_uleb128(self.sock)
        payload = recv(self.sock, num_bytes)

        log.xvc.debug(
            f"\tflags: {flags}, addr: {addr:X}, num_bytes: {num_bytes}, payload: {payload}"
        )

        if not self.pending_error:
            try:
                self.handler.write_mem(flags, addr, payload)
            except Exception as e:
                self.pending_error = str(e)
                log.xvc.error(self.pending_error)

        if not self.pending_error:
            log.xvc.debug("\tstatus: 0")
            self.sock.send(b"\x00")
        else:
            log.xvc.debug("\tstatus: 1")
            self.sock.send(b"\x01")

    def process_mrd(self):
        """
        XVC Format: [flags][addr][num bytes]

        Params:
        [flags] ULEB128 indicating flag bits (currently unused)
        [addr] ULEB128 starting address for mem read
        [num bytes] ULEB128 number of bytes in buffer

        Return Format: [buffer][status]
        [buffer] byte buffer of data read
        [status] error code of status (0x00 for success)
        """
        flags = get_uleb128(self.sock)
        addr = get_uleb128(self.sock)
        num_bytes = get_uleb128(self.sock)
        log.xvc.debug(f"\tflags: {flags}, addr: {addr:X}, num_bytes: {num_bytes}")

        data = b""
        if not self.pending_error:
            try:
                data = self.handler.read_mem(flags, addr, num_bytes)
            except Exception as e:
                self.pending_error = str(e)
                log.xvc.error(self.pending_error)

        if len(data) < num_bytes:
            data = bytes(num_bytes)
        self.sock.send(data)

        if self.pending_error:
            log.xvc.debug("\tstatus: 1")
            self.sock.send(b"\x01")
        else:
            log.xvc.debug("\tstatus: 0")
            self.sock.send(b"\x00")

    def process_stop(self):
        """
        Stops the XVC service on the server (used during testing).
        Only available when the "stop" capability is returned

        XVC Format:

        Return Format: [status]
        [status] error code of status (0x00 for success)
        """
        self.sock.send(b"\x00")
        self.running = False

    def process_default(self):
        self.pending_error = "Unhandled command"
        self.sock.send(b"\x01")

    def process(self, cmd):
        d = {
            "mwr": self.process_mwr,
            "mrd": self.process_mrd,
            "stop": self.process_stop,
            "getinfo": self.process_getinfo,
            "capabilities": self.process_capabilities,
            "settck": self.process_settck,
            "shift": self.process_shift,
            "error": self.process_error,
        }
        cmd = cmd.decode("utf-8")
        log.xvc.debug(f"{cmd}:")
        func = d.get(cmd, self.process_default)
        func()

    def run(self):
        self.running = True
        self.listen_sock = socket(AF_INET, SOCK_STREAM)
        server_address = ("", self.port)
        self.listen_sock.bind(server_address)
        self.listen_sock.listen(5)
        log.xvc.info(f"Starting service {server_address}")
        with self.listening:
            self.listening.notify()
        while self.running:
            try:
                self.sock, address = self.listen_sock.accept()
            except OSError as e:  # pragma: no cover
                if (
                    e.errno == 22 or e.errno == 10038
                ):  # invalid argument from socket already shuttdown
                    break
                else:
                    print(f"Unhandled OSError: {str(e)}, class {e.__class__}, dict {e.__dict__}")
                    raise e
            log.xvc.info("Accepted connection from {}".format(address))
            while self.running:
                try:
                    cmd = get_cmd(self.sock)
                    if cmd:
                        self.process(cmd)
                except Exception as e:  # pragma: no cover
                    log.xvc.error(f"{str(e)}")
                    break
            log.xvc.info("Connection {} disconnected".format(address))
            self.sock.close()
        self.listen_sock.close()
        log.xvc.info(f"Completed service {server_address}")

    def start(self, wait: int = 0):
        """
        Starts XVC server thread
        :param wait: Waits for listen socket to be listening before returning
        """
        self.thread = threading.Thread(target=self.run)
        self.thread.setDaemon(True)
        if wait:
            with self.listening:
                self.thread.start()
                self.listening.wait(wait)
        else:
            self.thread.start()

    def stop(self, wait: int = 5):
        """
        Stops XVC server thread and waits for it to close down.

        :param wait: timeout duration
        """
        self.running = False
        if self.sock:
            try:
                self.sock.close()
            except OSError:  # pragma: no cover
                pass
        if self.listen_sock:
            try:
                self.listen_sock.shutdown(SHUT_RD)
            except OSError:  # pragma: no cover
                pass
            self.listen_sock.close()
        if self.thread:
            self.thread.join(wait)
        self.thread = None


class XVCMemHandler(XVCHandler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.mem = dict()

    def get_mem(self, flags: int, addr: int, num_bytes: int) -> ByteString:
        data = bytearray()
        for n in range(num_bytes):
            data.append(self.mem.get(addr, 0))
            addr += 1
        return data

    def set_mem(self, flags: int, addr: int, data: ByteString):
        for d in data:
            self.mem[addr] = d
            addr += 1


class XVCVersalMemHandler(XVCMemHandler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_mem(0, 0xF1260114, bytes_from_words((0x02000000,)))  # npi ref clk enabled
        self.set_mem(0, 0xF1110884, bytes_from_words((0x1,)))  # done bit set



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from chipscopy.client.axis_trace_core_client import SERVICE_NAME
from chipscopy.tcf.services import DoneHWCommand, Service
from chipscopy.proxies.CorePropertyProxy import CorePropertyProxy


class AxisTraceService(Service):
    def getName(self) -> str:
        return SERVICE_NAME


class AxisTraceProxy(CorePropertyProxy, AxisTraceService):
    def __init__(self, channel):
        super().__init__(channel)
        self._listeners = {}

    #
    #  General methods
    #
    def initialize(self, node_id: str, done: DoneHWCommand) -> None:
        return self.send_xicom_command("initialize", (node_id,), done)

    def terminate(self, node_id: str, done: DoneHWCommand) -> None:
        return self.send_xicom_command("terminate", (node_id,), done)



# *****************************************************************************
# * Copyright (c) 2020 Xilinx, Inc.
# * Copyright (c) 2011, 2013 Wind River Systems, Inc. and others.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Wind River Systems - initial API and implementation
# *     Xilinx
# *****************************************************************************

import re
import time
from typing import ClassVar
from ... import channel, transport
from ...channel import fromJSONSequence, toJSONSequence
from .. import to_xargs, ServiceProvider, Service, add_service_provider


class CommandRequest(object):
    def __init__(self, _channel, token, data, node=None):
        self.channel = _channel
        self.token = token
        self.data = data
        if node is not None:
            self.node = node
        self.args = fromJSONSequence(data) if data else None

    def sendError(self, error, *result):
        error_props = {
            "Code": 0x20001,  # ERR_OTHER default used by hw_server
            "Time": int(time.perf_counter() * 1000),
            "Format": str(error),
            "Class": error.__class__.__name__,
            "Module": error.__class__.__module__
        }
        self.channel.sendResult(self.token, toJSONSequence((error_props, *result)))

    def sendResult(self, *result):
        self.channel.sendResult(self.token, toJSONSequence((None, *result)))

    def sendXicomResult(self, *results):
        result = to_xargs(results)
        self.channel.sendResult(self.token, toJSONSequence((None, *result)))

    def sendXicomProgress(self, *progress):
        args = to_xargs(progress)
        self.channel.sendProgress(self.token, toJSONSequence((None, *args)))


def send_xicom_event(service_name, event_name, args):
    data = toJSONSequence(to_xargs(args))
    transport.sendEvent(service_name, event_name, data)


def to_snake_case(name: str) -> str:
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


class CommandServer(channel.CommandServer):
    def __init__(self, service, _channel):
        self.service = service
        self.channel = _channel

    def command(self, token, name, data):
        cmd = getattr(self.service, to_snake_case(name) + "_cmd", None)
        if not cmd:
            cmd = getattr(self.service, name + 'Cmd', None)
        if cmd:
            try:
                request = CommandRequest(self.channel, token, data)
                cmd(request)
            except Exception as x:
                msg = f'ChipScope Service {self.service.getName()} {name}: {x}.'
                request.sendError(msg)
        else:
            channel.rejectCommand(token)


class LocalServiceProvider(ServiceProvider):
    def __init__(self, service):
        self.service = service

    def get_local_service(self, _channel):
        _channel.addCommandServer(
            self.service,
            CommandServer(self.service, _channel)
        )
        return self.service,


_local_services_added = set()


def add_local_service(service: Service or ClassVar[Service]):
    if str(service) not in _local_services_added:
        add_service_provider(LocalServiceProvider(service))
        _local_services_added.add(str(service))



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from chipscopy.client.core_property_client import CorePropertyClient

from chipscopy import dm
from chipscopy.client.core import CoreClient
from chipscopy.tcf.services import DoneHWCommand
from chipscopy.utils.logger import log


DOMAIN_NAME = "client_axis_pcie"

AXIS_PCIE_NODE_NAME = "pcie"
AXIS_PCIE_SERVICE_NAME = "AxisPCIe"


class AxisPCIeCoreClient(CorePropertyClient):
    @staticmethod
    def is_compatible(node: dm.Node) -> bool:
        return_val = node.type and node.type == AXIS_PCIE_NODE_NAME
        return return_val

    def get_service_proxy(self):
        return self.manager.channel.getRemoteService(AXIS_PCIE_SERVICE_NAME)

    def initialize(self, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        log[DOMAIN_NAME].info(f"Initializing {AXIS_PCIE_SERVICE_NAME} service")
        token = service.initialize(self.ctx, done_cb)
        return self.add_pending(token)

    def read_data(self, done: DoneHWCommand):
        service, done_cb = self.make_done(done)
        token = service.read_data(self.ctx, done_cb)
        return self.add_pending(token)

    def reset_core(self, done: DoneHWCommand):
        service, done_cb = self.make_done(done)
        token = service.reset_core(self.ctx, done_cb)
        return self.add_pending(token)

    # def read_data_region(self, int start_addr, int num_bytes, done: DoneHWCommand):
    #     service, done_cb = self.make_done(done)
    #     token = service.read_data_region(self.ctx, start_addr, num_bytes, done_cb)
    #     return self.add_pending(token)



# *****************************************************************************
# * Copyright (c) 2011, 2013, 2016 Wind River Systems, Inc. and others.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Wind River Systems - initial API and implementation
# *****************************************************************************

import threading
import types
from .. import protocol
from ..channel.Command import Command


class DispatchWrapper(object):
    "Simple wrapper for attribute access and invocation on TCF dispatch thread"
    def __init__(self, inner):
        self.inner = inner

    def __getattr__(self, attr):
        val = protocol.invokeAndWait(getattr, self.inner, attr)
        if type(val) in (types.FunctionType, types.MethodType):
            return DispatchWrapper(val)
        return val

    def __call__(self, *args, **kwargs):
        return protocol.invokeAndWait(self.inner, *args, **kwargs)


class CommandControl(object):
    """Provides a simple interface to send commands to remote services
    and receive results.

    Usage:
    > cmd = CommandControl(channel)
    > cmd.<service>.<command>(<args>)

    Examples:
    # send command, but don't wait for result:
    > cmd.RunControl.suspend("system")
    # getE() returns the result and raises an exception in case of error:
    > result = cmd.RunControl.getChildren(None).getE()
    # to get error and result at the same time, use this form:
    > error, result = cmd.Breakpoints.getIDs()
    """
    def __init__(self, channel, onDone=None, interactive=False):
        self._lock = threading.Condition()
        self._channel = channel
        self._onDone = onDone
        self._interactive = interactive
        self._queue = []
        self._pending = {}
        self._complete = []

    def __getattr__(self, attr):
        val = getattr(self._channel, attr, None)
        if val:
            if self._interactive and type(val) in (types.FunctionType,
                                                   types.MethodType):
                val = DispatchWrapper(val)
            return val
        services = protocol.invokeAndWait(self._channel.getRemoteServices)
        if attr == "services":
            return services
        if attr in services:
            return ServiceWrapper(self, attr)
        raise AttributeError("Unknown service: %s. Use one of %s" %
                             (attr, services))

    def invoke(self, service, command, *args, **kwargs):
        cmd = None
        if not protocol.isDispatchThread():
            if not kwargs.get("async"):
                cmd = protocol.invokeAndWait(self._invoke, service, command,
                                             *args, **kwargs)
                if cmd and self._interactive:
                    return cmd.getE()
            else:
                with self._lock:
                    self._queue.append((service, command, args, kwargs))
                    if len(self._queue) == 1:
                        protocol.invokeLater(self._processQueue)
                return
        return cmd

    def _invoke(self, service, command, *args, **kwargs):
        cmdCtrl = self

        class GenericCommand(Command):
            _result = None

            def done(self, error, args):  # @IgnorePep8
                resultArgs = None
                rcmds = ('read', 'readdir', 'roots')
                if not error and args:
                    # error result is usually in args[0],
                    # but there are exceptions
                    if service == "StackTrace" and command == "getContext":
                        error = self.toError(args[1])
                        resultArgs = (args[0],)
                    elif service == "Expressions" and command == "evaluate":
                        error = self.toError(args[1])
                        resultArgs = (args[0], args[2])
                    elif service == "FileSystem" and command in rcmds:
                        error = self.toError(args[1])
                        resultArgs = (args[0],) + tuple(args[2:])
                    elif service == "Streams" and command == 'read':
                        error = self.toError(args[1])
                        resultArgs = (args[0],) + tuple(args[2:])
                    elif service == "Diagnostics":
                        if command.startswith("echo"):
                            resultArgs = (args[0],)
                    else:
                        error = self.toError(args[0])
                        resultArgs = args[1:]
                cmdCtrl._doneCommand(self.token, error, resultArgs)

            def wait(self, timeout=None):
                cmdCtrl._waitForCommand(self.token, timeout)

            def cancel(self):
                return protocol.invokeAndWait(self.token.cancel)

            def getResult(self, wait=None):
                if wait:
                    cmdCtrl._waitForCommand(self.token)
                return self._result

            def getE(self):
                r = self.getResult(True)
                if r.error:
                    raise r.error
                return r.args

            def get(self):
                r = self.getResult(True)
                return r.args

            def getError(self):
                r = self.getResult(True)
                return r.error

            def __str__(self):
                if self._async:
                    return self.getCommandString()
                return str(self.get())

            def __iter__(self):
                return iter(self.getResult(True))
        cmd = GenericCommand(self._channel, service, command, args)
        cmd._async = kwargs.get("async")
        cmd._onDone = kwargs.get("onDone")
        self._addPending(cmd)
        return cmd

    def _processQueue(self):
        assert protocol.isDispatchThread()
        with self._lock:
            for cmd in self._queue:
                service, command, args, kwargs = cmd
                self._invoke(service, command, *args, **kwargs)
            del self._queue[:]

    def _addPending(self, cmd):
        with self._lock:
            self._pending[cmd.token.id] = cmd
            self._lock.notifyAll()

    def _doneCommand(self, token, error, args):
        with self._lock:
            cmd = self._pending.get(token.id)
            assert cmd
            del self._pending[token.id]
            cmd._result = CommandResult(token, error, args)
            if cmd._async:
                self._complete.append(cmd)
            isDone = self.isDone()
            if isDone:
                self._lock.notifyAll()
        if cmd._onDone:
            if args is None:
                args = (None,)
            cmd._onDone(error, *args)
        if isDone and self._onDone:
            self._onDone()

    def isDone(self):
        with self._lock:
            return not self._pending and not self._queue

    def wait(self, timeout=None):
        assert not protocol.isDispatchThread()
        with self._lock:
            while self._pending or self._queue:
                self._lock.wait(timeout)
                if timeout:
                    break

    def _waitForCommand(self, token, timeout=None):
        assert not protocol.isDispatchThread()
        with self._lock:
            while token.id in self._pending:
                self._lock.wait(timeout)
                if timeout:
                    break
            else:
                if self._queue:
                    self._lock.wait(timeout)
                    while token.id in self._pending:
                        self._lock.wait(timeout)
                        if timeout:
                            break

    def cancel(self):
        if not protocol.isDispatchThread():
            protocol.invokeLater(self.cancel)
            return
        with self._lock:
            for cmd in list(self._pending.values()):
                cmd.token.cancel()
            del self._queue[:]

    def getResult(self, wait=True):
        if wait:
            self.wait()
        with self._lock:
            result = [c.getResult() for c in self._complete]
            del self._complete[:]
        return result


class CommandResult(object):
    def __init__(self, token, error, args):
        self.token = token
        self.error = error
        # unwrap result if only one element
        if args and len(args) == 1:
            args = args[0]
        self.args = args

    def __str__(self):
        if self.error:
            return "[%s] error: %s" % (self.token.id, self.error)
        return "[%s] result: %s" % (self.token.id, self.args)
    __repr__ = __str__

    def __iter__(self):
        yield self.error
        yield self.args


class ServiceWrapper(object):
    def __init__(self, control, service):
        self._control = control
        self._service = service

    def __getattr__(self, attr):
        return CommandWrapper(self._control, self._service, attr)


class CommandWrapper(object):
    def __init__(self, control, service, command):
        self._control = control
        self._service = service
        self._command = command

    def __call__(self, *args, **kwargs):
        return self._control.invoke(self._service, self._command, *args,
                                    **kwargs)



# *****************************************************************************
# * Copyright (c) 2020 Xilinx, Inc.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Xilinx
# *****************************************************************************

import socket
import time


DISCOVERY_TCF_PORT = 1534
TCF_VERSION = '2'
UDP_REQ_INFO = 1
UDP_PEER_INFO = 2
UDP_REQ_SLAVES = 3
UDP_SLAVES_INFO = 4
UDP_PEERS_REMOVED = 5


class ServerPeer(object):
    def __init__(self, address, name):
        self.address = address
        self.name = name

    def get_address(self):
        return self.address

    def get_name(self):
        return self.name

    def __str__(self):
        return '{}, {}'.format(self.address, self.name)


NAME_ATTR = 'Name'
TRANSPORT_ATTR = 'TransportName'
HOST_ATTR = 'Host'
PORT_ATTR = 'Port'


def find_peers(timeout=5):
    peers = []

    header = bytearray([ord('T'), ord('C'), ord('F'), ord(TCF_VERSION), UDP_REQ_INFO, 0, 0, 0])
    bcast_addr = '<broadcast>'
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
    sock.bind(('', 0))  # use local ip address and random port

    sock.sendto(bytes(header), (bcast_addr, DISCOVERY_TCF_PORT))

    peer_packets = []
    t_end = time.time() + timeout
    print('Searching for servers for {} seconds.  Please wait...'.format(timeout))
    while time.time() < t_end:
        data, address = sock.recvfrom(1024)
        recv_header = list(data[:8].decode('utf-8'))
        req_type = ord(recv_header[4])
        recv_data = data[8:].decode('utf-8')
        if req_type == UDP_PEER_INFO and len(recv_data) > 0:
            peer_packets.append(recv_data)

    for packet in peer_packets:
        peer_attrs = packet.split('\0')
        peer_name = ''
        peer_transport = ''
        peer_host = ''
        peer_port = ''
        for attr in peer_attrs:
            if '=' not in attr:
                continue
            key, value = attr.split('=', 1)
            if key == NAME_ATTR:
                peer_name = value
            elif key == TRANSPORT_ATTR:
                peer_transport = value
            elif key == HOST_ATTR:
                peer_host = value
            elif key == PORT_ATTR:
                peer_port = value
        peer_addr = peer_transport + ':' + peer_host + ':' + peer_port
        dup_peer = False
        for peer in peers:
            if peer.get_address() == peer_addr and peer.get_name() == peer_name:
                dup_peer = True
                break
        if not dup_peer:
            peers.append(ServerPeer(peer_addr, peer_name))

    return peers

